
"""
Contains system specifc information. At the Moment, this is all hard-coded.
In the future, this will be part of the installation of AIzymes.

set_system() contains general variables.
submit_head() constructs the submission header to submit jobs.

"""

import sys
import os
import getpass
import json
from time import sleep

def set_system(self):

    standard_settings = {
        "rosetta_ext"          : "linuxgccrelease",
        "bash_args"            : "",
        "ROSETTA_PATH"         : f"/u/bunzela/bin/rosetta.source.release-371/main/source/",
        "FIELD_TOOLS"          : f'/u/{getpass.getuser()}/bin/AIzymes_dev/src/aizymes/FieldTools.py',
        "FOLDER_ProteinMPNN"   : f'/u/{getpass.getuser()}/bin/ProteinMPNN',
        "FOLDER_ProteinMPNN_h" : f'/u/{getpass.getuser()}/bin/ProteinMPNN/helper_scripts',
        "FOLDER_LigandMPNN"    : f'/u/{getpass.getuser()}/bin/LigandMPNN',
        "FOLDER_Alphafold"     : f'/u/bunzela/bin/alphafold',
        "SUBMIT_HEAD"          : submit_head(self),
    }

    if os.path.isfile(self.SYSTEM):

        # Load settings from config
        with open(self.SYSTEM, 'r') as f:
            loaded_settings = json.load(f)
        
        # Assign to self
        for key, value in loaded_settings.items():
            setattr(self, key, value)
            
    else:

        # MAke sure config file is not save in a path that is deleted during AI.zymes setup
        system_path = os.path.abspath(self.SYSTEM)
        current_dir = os.path.abspath(self.FOLDER_HOME)
        if os.path.commonpath([system_path, current_dir]) == current_dir:
            print(f"Error! {self.SYSTEM} cannot be in current directory or subdirectory thereof")
            sys.exit()
        
        print(f"Config file {self.SYSTEM} missing. Lets create a new config file!\n")

        default = standard_settings["ROSETTA_PATH"]
        self.ROSETTA_PATH = input(f"Path to Rosetta [leavel blank for default: {default}]") or default
        print(f"ROSETTA_PATH set to {self.ROSETTA_PATH}\n")
        if not os.path.isdir(self.ROSETTA_PATH):
            print(f"ERROR, ROSETTA_PATH {self.ROSETTA_PATH} does not exist!")
            sys.exit()
            
        default = standard_settings["FIELD_TOOLS"]
        self.FIELD_TOOLS = input(f"Path to FieldTools [leavel blank for default: {default}]") or default
        print(f"FIELD_TOOLS set to {self.FIELD_TOOLS}\n")
        if not os.path.isfile(self.FIELD_TOOLS):
            print(f"ERROR, FIELD_TOOLS {self.FIELD_TOOLS} does not exist!")
            sys.exit()
        
        default = standard_settings["FOLDER_ProteinMPNN"]
        self.FOLDER_ProteinMPNN = input(f"Path to ProteinMPNN [leavel blank for default: {default}]") or default
        print(f"ProteinMPNN set to {self.FOLDER_ProteinMPNN}\n")
        if not os.path.isdir(self.FOLDER_ProteinMPNN):
            print(f"ERROR, FOLDER_ProteinMPNN {self.FOLDER_ProteinMPNN} does not exist!")
            sys.exit()

        self.FOLDER_ProteinMPNN_h = f'{self.FOLDER_ProteinMPNN}/helper_scripts'
        if not os.path.isdir(self.FOLDER_ProteinMPNN_h):
            print(f"ERROR, FOLDER_ProteinMPNN_h {self.FOLDER_ProteinMPNN_h} does not exist!")
            sys.exit()

        default = standard_settings["FOLDER_LigandMPNN"]
        self.FOLDER_LigandMPNN = input(f"Path to LigandMPNN [leavel blank for default: {default}]") or default
        print(f"FOLDER_LigandMPNN set to {self.FOLDER_LigandMPNN}\n")
        if not os.path.isdir(self.FOLDER_LigandMPNN):
            print(f"ERROR, FOLDER_LigandMPNN {self.FOLDER_LigandMPNN} does not exist!")
            sys.exit()

        default = standard_settings["FOLDER_Alphafold"]
        self.FOLDER_Alphafold = input(f"Path to Alphafold [leavel blank for default: {default}]") or default
        print(f"FOLDER_ALPHAFOLD set to {self.FOLDER_Alphafold}\n")
        if not os.path.isdir(self.FOLDER_Alphafold):
            print(f"ERROR, FOLDER_ALPHAFOLD {self.FOLDER_Alphafold} does not exist!")
            sys.exit()

        default = standard_settings["rosetta_ext"]
        self.rosetta_ext = input(f"Extension for installed Rosetta Programms [leavel blank for default: {default}]") or default
        print(f"rosetta_ext set to {self.rosetta_ext}\n")

        self.bash_args = standard_settings["bash_args"]
        
        default = standard_settings["SUBMIT_HEAD"]
        self.SUBMIT_HEAD = input(f"""Path to controller submission header.
Do not include --job-name, --output, and --error flags!
[leave blank for default header]""") or default
        if self.SUBMIT_HEAD != default:
            if not os.path.isfile(self.SUBMIT_HEAD):
                print(f"ERROR, SUBMIT_HEAD {self.SUBMIT_HEAD} does not exist!")
                sys.exit()
            with open(self.SUBMIT_HEAD, "r") as f:
                self.SUBMIT_HEAD = f.read()
        print(f"\nSUBMIT_HEAD set to\n{self.SUBMIT_HEAD}\n")
        
        # Save settings
        standard_settings = {
            "rosetta_ext"          : self.rosetta_ext,
            "bash_args"            : self.bash_args,
            "ROSETTA_PATH"         : self.ROSETTA_PATH,
            "FIELD_TOOLS"          : self.FIELD_TOOLS,
            "FOLDER_ProteinMPNN"   : self.FOLDER_ProteinMPNN,
            "FOLDER_ProteinMPNN_h" : self.FOLDER_ProteinMPNN_h,
            "FOLDER_LigandMPNN"    : self.FOLDER_LigandMPNN,
            "FOLDER_Alphafold"     : self.FOLDER_Alphafold,
            "SUBMIT_HEAD"          : self.SUBMIT_HEAD
        }
        with open(self.SYSTEM, 'w') as f:
            json.dump(standard_settings, f, indent=4) 
        print(f"Config file saved to {self.SYSTEM}\n")
        print(f"TIP: To avoid making a new config file everytime, store path to config file in SYSTEM during AI.zymes setup")
        print()        
        sleep(1.0)

def submit_head(self):

    
    if self.MAX_GPUS == 0: 
        partitition='general'
    else:
        partitition='gpu'
    
    cmd = f"""#!/bin/bash
#SBATCH --partition={partitition}
#SBATCH --nodes=1
#SBATCH --ntasks=1
#SBATCH --cpus-per-task={self.MAX_JOBS}
#SBATCH --mem={self.MEMORY}G
#SBATCH --time=15:00:00
"""
    if self.MAX_GPUS > 0:
        cmd += f"""#SBATCH --gres=gpu:{self.MAX_GPUS}
"""

    return cmd


import os
import logging
import sys
import shlex
import shutil

# Import helper functions (ensure helper_002 is in PYTHONPATH)
from helper_002 import sequence_from_pdb as _sequence_from_pdb, generate_remark_from_all_scores_df  # type: ignore


def _generate_chai1_fasta(self, index, fasta_path):
    """
    Generates the FASTA input file for Chai-1 based on AIzymes config.
    Handles single protein sequence retrieval from a .seq file, with a fallback to .pdb if needed.
    Expects self.LIGAND to be a SMILES string.
    Returns a tuple: (success: bool, protein_sequence_length: int or None)
    """
    logging.debug(f"Generating Chai-1 FASTA for index {index} at {fasta_path}")

    # Get protein input base from the scores DataFrame
    protein_input_base = self.all_scores_df.at[int(index), "step_input_variant"]
    protein_seq_file = f"{protein_input_base}.seq"
    protein_pdb_file = f"{protein_input_base}.pdb"

    protein_sequence = None
    if os.path.exists(protein_seq_file):
        logging.debug(f"Reading sequence from {protein_seq_file}")
        try:
            with open(protein_seq_file, 'r') as f:
                protein_sequence = f.read().strip()
        except Exception as e:
            logging.error(f"Error reading {protein_seq_file}: {e}")
    elif os.path.exists(protein_pdb_file):
        logging.warning(f"Sequence file {protein_seq_file} not found. Extracting from {protein_pdb_file}.")
        protein_sequence = _sequence_from_pdb(protein_input_base)
    else:
        logging.error(f"Cannot find sequence source for index {index}: neither {protein_seq_file} nor {protein_pdb_file} exists.")
        return False, None

    if not protein_sequence or any(x in protein_sequence for x in ["NOTFOUND", "ERROR", "NOTPARSED"]):
        logging.error(f"Invalid protein sequence for index {index}. Obtained: '{protein_sequence}'")
        return False, None

    # Retrieve ligand identifier: use CHAI1_LIGAND_IDENTIFIER if set, otherwise fallback to LIGAND
    ligand_id = getattr(self, 'CHAI1_LIGAND_IDENTIFIER', None)
    if not ligand_id:
        ligand_id = getattr(self, 'LIGAND', None)
    if not ligand_id or not isinstance(ligand_id, str):
        logging.error(f"Index {index}: Neither CHAI1_LIGAND_IDENTIFIER nor LIGAND is a valid string.")
        return False, None

    # Write FASTA with two entries: protein and ligand
    try:
        with open(fasta_path, 'w') as f:
            f.write(">protein|name=ChainA\n")
            f.write(f"{protein_sequence}\n")
            f.write(">ligand|name=ChainX\n")
            f.write(f"{ligand_id}\n")
        logging.info(f"Successfully generated Chai-1 FASTA at {fasta_path}")
        return True, len(protein_sequence)
    except IOError as e:
        logging.error(f"Failed to write FASTA file {fasta_path}: {e}")
        return False, None


def prepare_Chai1(self, index, cmd, gpu_id=None):
    """
    Prepares the shell command for running Chai-1 prediction.
    Sets up the working directory, generates the FASTA input, constructs the chai-lab command with options,
    and appends post-processing steps for converting the output CIF to PDB with added REMARKs.
    """
    logging.info(f"Preparing Chai-1 for index {index}.")

    # Check required attributes
    required_attrs = ['FOLDER_DESIGN', 'WT', 'all_scores_df', 'LIGAND']
    missing_attrs = [attr for attr in required_attrs if not hasattr(self, attr) or getattr(self, attr) is None]
    if missing_attrs:
        logging.error(f"Missing required attributes for Chai-1 index {index}: {', '.join(missing_attrs)}")
        return cmd + f"\n# ERROR: Missing attributes for Chai-1: {', '.join(missing_attrs)}\n"

    try:
        index_str = str(index)
        folder_design = self.FOLDER_DESIGN
        wt_name = self.WT
        
        # Define working and output directories
        chai1_work_dir_py = os.path.join(folder_design, index_str, "Chai1")
        aizymes_index_dir_py = os.path.join(folder_design, index_str)
        os.makedirs(chai1_work_dir_py, exist_ok=True)

        # Define separate output directory inside working directory
        chai1_output_dir_py = os.path.join(chai1_work_dir_py, "output")
        if os.path.exists(chai1_output_dir_py):
            shutil.rmtree(chai1_output_dir_py)
        os.makedirs(chai1_output_dir_py)
        # Path for input FASTA
        fasta_input_path_py = os.path.join(chai1_work_dir_py, f"{wt_name}_Chai1_{index_str}.fasta")

        # Shell-quoting for safety
        chai1_work_dir_sh = shlex.quote(chai1_work_dir_py)
        fasta_input_path_sh = shlex.quote(fasta_input_path_py)
        chai1_output_dir_sh = shlex.quote(chai1_output_dir_py)
    except Exception as e:
        logging.error(f"Error defining paths for index {index}: {e}")
        return cmd + f"\n# ERROR: Path definition failed for Chai-1 index {index}: {e}\n"

    # Generate FASTA input
    fasta_success, protein_seq_len = _generate_chai1_fasta(self, index, fasta_input_path_py)
    if not fasta_success:
        logging.error(f"Failed to generate FASTA for Chai-1 index {index}.")
        return cmd + f"\n# ERROR: Failed to generate Chai-1 FASTA for index {index}.\n"

    try:
        # Calculate ligand residue number for later renumbering (protein length + 1)
        ligand_resi_py = (protein_seq_len + 1) if (protein_seq_len and isinstance(protein_seq_len, int)) else 9999
        # Determine ligand name from config (prefer CHAI1_LIGAND_IDENTIFIER, fallback to LIGAND, default 'LIG')
        ligand_name_py = getattr(self, 'LIGAND', 'LIG')
        ligand_name_py = str(ligand_name_py) if ligand_name_py else 'LIG'

        # Initialize command accumulator for this Chai-1 job
        current_cmd = f"\n### Chai-1 Prediction ###\n"

        # Activate specific environment if defined
        chai1_env_activate = getattr(self, 'CHAI1_ENV_ACTIVATE', None)
        if chai1_env_activate:
            current_cmd += f"# Activating Chai-1 environment\n{chai1_env_activate}\n"

        # Set GPU if provided
        if gpu_id is not None:
            current_cmd += f"export CUDA_VISIBLE_DEVICES={gpu_id}\n"
            logging.info(f"Chai-1 index {index} assigned to GPU: {gpu_id}")
        else:
            logging.info(f"Chai-1 index {index} running with default GPU/CPU settings.")

        # Build command parts
        chai1_cmd_parts = [
            "chai-lab", "fold", fasta_input_path_sh, chai1_output_dir_sh
        ]

        # Optional parameters
        num_samples = getattr(self, 'CHAI1_DIFFUSION_SAMPLES', 1)
        chai1_cmd_parts.extend(["--num-diffn-samples", str(num_samples)])

        num_recycles = getattr(self, 'CHAI1_TRUNK_RECYCLES', 3)
        chai1_cmd_parts.extend(["--num-trunk-recycles", str(num_recycles)])

        seed = getattr(self, 'SEED', None)
        if seed is not None:
            chai1_cmd_parts.extend(["--seed", str(seed + int(index))])

        if getattr(self, 'CHAI1_USE_MSA_SERVER', False):
            chai1_cmd_parts.append("--use-msa-server")
            msa_server_url = getattr(self, 'CHAI1_MSA_SERVER_URL', None)
            if msa_server_url:
                chai1_cmd_parts.extend(["--msa-server-url", shlex.quote(msa_server_url)])

        if getattr(self, 'CHAI1_USE_TEMPLATES_SERVER', False):
            chai1_cmd_parts.append("--use-templates-server")

        if getattr(self, 'CHAI1_LOW_MEMORY', True):
            chai1_cmd_parts.append("--low-memory")

        if getattr(self, 'CST_NAME', None) is not None:
            cst_file_py = os.path.join(self.FOLDER_PARENT, f"{self.CST_NAME}.csv")
            if os.path.exists(cst_file_py):
                chai1_cmd_parts.extend(["--constraint-path", shlex.quote(cst_file_py)])
                logging.info(f"Using constraints file for Chai-1: {cst_file_py}")
            else:
                logging.warning(f"Constraint file {cst_file_py} specified but not found for Chai-1.")

        current_cmd += shlex.join(chai1_cmd_parts) + "\n\n"

        # AWK scripts for ligand processing (primary: chain B resi 1, fallback: any LIG)
        # Using raw strings to avoid escaping issues
        awk_script_primary_py = r"""
BEGIN { if (new_resi+0 != new_resi) { new_resi=9999 } }
( ($1=="ATOM"||$1=="HETATM") && substr($0,18,3)=="LIG" && substr($0,22,1)=="B" && substr($0,23,4)+0==1 ) {
    serial   = substr($0,7,5); atom_name = substr($0,13,4); alt_loc   = substr($0,17,1);
    icode    = substr($0,27,1); x         = substr($0,31,8); y         = substr($0,39,8);
    z        = substr($0,47,8); occ       = substr($0,55,6); temp      = substr($0,61,6);
    element  = substr($0,77,2); charge    = substr($0,79,2);
    sub(/_.*/, "", atom_name);
    printf "ATOM  %5s %-4s%1s%3s %1s%4d%1s   %8s%8s%8s%6s%6s          %2s%2s\\n", \
           serial, atom_name, alt_loc, new_resn, new_chain, new_resi, icode, x, y, z, occ, temp, element, charge;
    found = 1
}
END { exit !found }
"""
        awk_script_fallback_py = r"""
BEGIN { if (new_resi+0 != new_resi) { new_resi=9999 } }
( ($1=="ATOM"||$1=="HETATM") && substr($0,18,3)=="LIG" ) {
    serial   = substr($0,7,5); atom_name = substr($0,13,4); alt_loc   = substr($0,17,1);
    icode    = substr($0,27,1); x         = substr($0,31,8); y         = substr($0,39,8);
    z        = substr($0,47,8); occ       = substr($0,55,6); temp      = substr($0,61,6);
    element  = substr($0,77,2); charge    = substr($0,79,2);
    sub(/_.*/, "", atom_name);
    printf "ATOM  %5s %-4s%1s%3s %1s%4d%1s   %8s%8s%8s%6s%6s          %2s%2s\\n", \
           serial, atom_name, alt_loc, new_resn, new_chain, new_resi, icode, x, y, z, occ, temp, element, charge;
    found = 1
}
END { exit !found }
"""

        # Generate REMARKs
        remarks_py = ""
        try:
            remarks_py = generate_remark_from_all_scores_df(self, index=index)
            logging.info(f"Generated REMARKs for Chai-1 index {index}")
        except Exception as e:
            logging.error(f"Error generating remarks for Chai-1 index {index}: {e}")
            remarks_py = "REMARK 999 Error generating remarks\n"

        # Define expected output files for post-processing
        model_idx = 0
        expected_output_cif_py = os.path.join(chai1_output_dir_py, f"pred.model_idx_{model_idx}.cif")
        expected_output_cif_sh = shlex.quote(expected_output_cif_py)
        expected_scores_npz_py = os.path.join(chai1_output_dir_py, f"scores.model_idx_{model_idx}.npz")
        expected_scores_npz_sh = shlex.quote(expected_scores_npz_py)

        final_structure_aizymes_name_py = os.path.join(aizymes_index_dir_py, f"{wt_name}_Chai1_{index_str}.pdb")
        final_structure_aizymes_name_sh = shlex.quote(final_structure_aizymes_name_py)
        final_scores_aizymes_name_py = os.path.join(aizymes_index_dir_py, f"scores_Chai1_{index_str}.npz")
        final_scores_aizymes_name_sh = shlex.quote(final_scores_aizymes_name_py)

        # Post-processing command
        post_process_cmd = f"""
# Post-process Chai-1 output
echo "Checking for Chai-1 output CIF: {expected_output_cif_sh}"
if [ -f {expected_output_cif_sh} ] && [ -s {expected_output_cif_sh} ]; then
  echo "Chai-1 output found. Converting CIF to PDB and adding remarks..."

  TMP_PDB_CONVERTED=$(mktemp)
  TMP_PDB_FINAL=$(mktemp)

  # 1. Convert CIF to PDB using the helper script
  echo "Running cif_to_pdb.py..."
  python "{self.FOLDER_PARENT}/cif_to_pdb.py" --cif_file {expected_output_cif_sh} --pdb_file "$TMP_PDB_CONVERTED"
  if [ $? -ne 0 ] || [ ! -s "$TMP_PDB_CONVERTED" ]; then
      echo "ERROR: cif_to_pdb.py conversion failed or produced empty file for {expected_output_cif_sh}" >&2
      rm -f "$TMP_PDB_CONVERTED" "$TMP_PDB_FINAL"
      exit 1
  fi
  echo "CIF converted to $TMP_PDB_CONVERTED"
  # Remove any TER records and hydrogens to clean the converted PDB
  sed -i '/TER/d' "$TMP_PDB_CONVERTED"
  sed -i '/ H  /d' "$TMP_PDB_CONVERTED"

  #############################
  # Standardise ligand records #
  #############################
  # Variables passed in from Python for ligand renaming/renumbering (already shell-safe)
  LIGAND_NAME=$(printf '%s' {shlex.quote(ligand_name_py)})
  LIGAND_RESI=$(printf '%s' {shlex.quote(str(ligand_resi_py))})
  REMARKS=$(printf '%s' {shlex.quote(remarks_py)})

  # 1. Start final PDB file with remarks header (match Boltz/ESMFold format)
  echo "REMARK   0 File processed from Chai-1 prediction" > "$TMP_PDB_FINAL"
  echo "REMARK   0 Chain A: Protein, Chain X: Ligand ($LIGAND_NAME)" >> "$TMP_PDB_FINAL"
  echo "$REMARKS" >> "$TMP_PDB_FINAL"

  # 2. Append protein ATOM lines (excluding original ligand) from converted PDB
  grep '^ATOM  ' "$TMP_PDB_CONVERTED" | grep -v " LIG " >> "$TMP_PDB_FINAL" || echo "Warning: No protein ATOM lines found or grep failed."
  echo "TER" >> "$TMP_PDB_FINAL"

  # 3. Process ligand lines via two-phase AWK (primary then fallback)
  LIGAND_LINES=$(mktemp)
  awk -v new_resn="$LIGAND_NAME" -v new_chain="X" -v new_resi="$LIGAND_RESI" \
      '{awk_script_primary_py}' \
      "$TMP_PDB_CONVERTED" > "$LIGAND_LINES"
  if [ $? -eq 0 ] && [ -s "$LIGAND_LINES" ]; then
      cat "$LIGAND_LINES" >> "$TMP_PDB_FINAL"
  else
      echo "Warning: primary ligand AWK failed, trying fallback..." >&2
      awk -v new_resn="$LIGAND_NAME" -v new_chain="X" -v new_resi="$LIGAND_RESI" \
          '{awk_script_fallback_py}' \
          "$TMP_PDB_CONVERTED" >> "$TMP_PDB_FINAL"
  fi
  rm -f "$LIGAND_LINES"
  echo "TER" >> "$TMP_PDB_FINAL"

  # 4. Finish file
  echo "END" >> "$TMP_PDB_FINAL"

  # 5. Move final PDB
  if [ -s "$TMP_PDB_FINAL" ]; then
      echo "Moving final PDB to {final_structure_aizymes_name_sh}"
      mv "$TMP_PDB_FINAL" {final_structure_aizymes_name_sh}
  else
      echo "ERROR: Final PDB assembly failed (empty file)." >&2
      rm -f "$TMP_PDB_CONVERTED" "$TMP_PDB_FINAL"
      exit 1
  fi
  rm -f "$TMP_PDB_CONVERTED"

  # 6. Copy score file
  if [ -f {expected_scores_npz_sh} ]; then
      echo "Copying Chai-1 scores NPZ to {final_scores_aizymes_name_sh}"
      cp {expected_scores_npz_sh} {final_scores_aizymes_name_sh}
  else
      echo "Warning: Chai-1 scores file not found: {expected_scores_npz_sh}" >&2
  fi
else
  echo "ERROR: Expected Chai-1 output CIF not found or empty: {expected_output_cif_sh}" >&2
  echo "Contents of Chai-1 output directory ({chai1_work_dir_sh}):"
  ls -la {chai1_work_dir_sh} || echo "Cannot access directory"
  exit 1
fi
"""
        
        current_cmd += post_process_cmd

        # Optional cleanup
        if getattr(self, 'REMOVE_TMP_FILES', False):
             current_cmd += f"\n# Removing temporary Chai-1 directory\necho \"Removing temporary directory: {chai1_work_dir_sh}\"\nrm -r {chai1_work_dir_sh}\n"

        logging.info(f"Chai-1 command for index {index} prepared successfully.")
        return cmd + current_cmd.strip() + "\n"
    except Exception as e:
        logging.error(f"Error preparing Chai-1 for index {index}: {e}")
        return cmd + f"\n# ERROR: Failed to prepare Chai-1 for index {index}: {e}\n" 


def plot_tree(self, 
              max_generation=0, 
              radius_offset=0, 
              max_angle=360, 
              color="combined_score",
              generation_labels=None, 
              start_angle_label=0, 
              filename=None):

    import os
    import sys
    import pandas as pd
    import matplotlib.pyplot as plt
    import numpy as np
    import networkx as nx
    import math

    from matplotlib.colors import ListedColormap, BoundaryNorm, TwoSlopeNorm, PowerNorm
    from matplotlib.colors import LinearSegmentedColormap, Normalize

    if max_generation == 0:
        max_generation = int(self.plot_scores_df['generation'].max() + 1)

    if "redox" not in self.SELECTED_SCORES:
        self.TARGET_REDOX=0
        
    # Create a directed graph
    G = nx.DiGraph()
    G.add_node(0, 
               sequence="X",
               generation=0,
               hamming_distance=np.nan,
               BioDC_redox=np.nan,
               **{f"{score}_score": np.nan for score in self.SELECTED_SCORES}, 
               final_score=np.nan, 
               combined_score=np.nan)
    
    # Build a mapping from the original DataFrame index to a new sequential index (starting at 1)
    index_mapping = {old_idx: new_idx for new_idx, old_idx in enumerate(self.plot_scores_df.index, start=1)}
    
    # Iterate over the rows using iterrows() to get both the original index and the row data
    for old_idx, row in self.plot_scores_df.iterrows():
        new_idx = index_mapping[old_idx]  # new sequential row number
    
        if pd.isna(row['parent_index']):
            continue
        if row['generation'] + 1 > max_generation and max_generation != 0:
            continue
    
        # Adjust the parent's index using the mapping.
        if row['parent_index'] != "Parent":
            parent_old = int(float(row['parent_index']))
            parent_idx = index_mapping.get(parent_old, 0)
        else:
            parent_idx = 0
    
        distance = row['parent_mutations']
        if not isinstance(distance, (int, float)) or np.isnan(distance):
            print(f"Illegal hamming_distance value detected: {distance} for variant {new_idx}")
            sys.exit()

        if "redox" in self.SELECTED_SCORES:
            BioDC_redox = row['BioDC_redox']
        else:
            BioDC_redox = np.nan            

        G.add_node(new_idx,
                   sequence            = row['sequence'],
                   generation          = row['generation'] + 1,
                   hamming_distance    = distance,
                   BioDC_redox         = abs(BioDC_redox-self.TARGET_REDOX),
                   **{f"{score}_score" : self.scores[f"{score}_score"][new_idx - 1] for score in self.SELECTED_SCORES},
                   final_score         = self.scores[f"final_score"][new_idx - 1],
                   combined_score      = self.scores['combined_score'][new_idx - 1])
    
        G.add_edge(parent_idx,
                   new_idx,
                   hamming_distance    = distance,
                   BioDC_redox         = abs(BioDC_redox-self.TARGET_REDOX),
                   **{f"{score}_score" : self.scores[f"{score}_score"][new_idx - 1] for score in self.SELECTED_SCORES},
                   final_score         = self.scores[f"final_score"][new_idx - 1],
                   combined_score      = self.scores['combined_score'][new_idx - 1])

    # Calculate subtree sizes
    def calculate_subtree_sizes(graph, node, subtree_sizes):
        size = 1
        for child in graph.successors(node):
            size += calculate_subtree_sizes(graph, child, subtree_sizes)
        subtree_sizes[node] = size
        return size

    subtree_sizes = {}
    calculate_subtree_sizes(G, 0, subtree_sizes)

    # Assign polar coordinates with grouping
    pos_polar = {}

    def assign_polar_positions(node, radius=0, angle_start=0, angle_range=max_angle):
        pos_polar[node] = (radius, math.radians(angle_start))
        children = list(G.successors(node))
        if children:
            children.sort(key=lambda x: subtree_sizes[x])
            num_children = len(children)
            child_angle_range = angle_range / num_children
            for i, child in enumerate(children):
                child_angle_start = angle_start + i * child_angle_range
                assign_polar_positions(child, radius + 1, child_angle_start, child_angle_range)

    assign_polar_positions(0, radius=0, angle_start=0, angle_range=max_angle)

    def adjust_uniform_angles(pos_polar):
        unique_angles = sorted({math.degrees(angle) for _, angle in pos_polar.values()})
        num_angles = len(unique_angles)
        uniform_angles = [max_angle / num_angles * i for i in range(num_angles)]
        angle_mapping = {unique_angles[i]: uniform_angles[i] for i in range(num_angles)}
        for node, (radius, angle) in pos_polar.items():
            pos_polar[node] = (radius, math.radians(angle_mapping[math.degrees(angle)]))
    adjust_uniform_angles(pos_polar)

    for node, (radius, angle) in pos_polar.items():
        if radius != 0:
            pos_polar[node] = (radius + radius_offset, angle)

    def polar_to_cartesian(radius, angle):
        return radius * math.cos(angle), radius * math.sin(angle)

    # Set up the colormap and normalization
    if color == "hamming_distance":
        min_color = 0
        max_color = int(max([G.edges[edge]['hamming_distance'] for edge in G.edges]))
        base_cmap = LinearSegmentedColormap.from_list("WhiteToBlue", [(1, 1, 1), (0, 0, 1)])
        bounds = np.linspace(min_color, max_color + 1, max_color + 1)
        norm = BoundaryNorm(bounds, ncolors=len(bounds) - 1, clip=True)
        custom_cmap = ['salmon'] + [base_cmap(i / max_color) for i in range(1, max_color)]
        custom_cmap = ListedColormap(custom_cmap)
        norm = BoundaryNorm(bounds, ncolors=len(bounds) - 1, clip=True)
        color_values = [G.edges[edge]['hamming_distance'] for edge in G.edges]
    else:
        values = [G.edges[edge][color] for edge in G.edges]

        if color ==  "BioDC_redox":

            """
            vmin=-1
            vmax=1
            vmid=self.HIGHSCORE["BioDC_redox"]
            base_cmap = LinearSegmentedColormap.from_list(
                "RedYellowBlue",
                [
                    (0.0, (1, 0, 0)),
                    (0.3, (1, 0.5, 0)),
                    (0.5, (1, 1, 0)),
                    (0.7, (0.5, 0.5, 1)),
                    (1.0, (0, 0, 1))
                ]
            )
            norm = TwoSlopeNorm(vmin=vmin, vcenter=vmid, vmax=vmax)   
            """

            #vmax = abs(np.amax(self.scores['BioDC_redox'] - self.TARGET_REDOX))
            #norm = PowerNorm(gamma=0.5, vmin=0, vmax=vmax)
            #base_cmap = LinearSegmentedColormap.from_list("BlueToWhie", [(0, 0, 1), (1, 1, 1)])
            
            # Grab your edge values
            color_values = [G.edges[edge][color] for edge in G.edges]
            
            # Pride flag colors
            pride_colors = [
                "#750787",  # violet
                "#004dff",  # blue
                "#008026",  # green
                "#ffed00",  # yellow
                "#ff8c00",  # orange
                "#e40303"   # red
            ]
            n_colors = len(pride_colors)
            
            # Compute max value for bounds
            vmax = max(color_values)
            boundaries = np.linspace(0, vmax, n_colors + 1)

            # clip at 0.3
            vmax = 0.3
            boundaries = np.linspace(0, vmax, n_colors + 1)
            color_values = [min(val, vmax) for val in color_values]

            # Set colormap and normalization
            base_cmap = ListedColormap(pride_colors)
            norm = BoundaryNorm(boundaries, base_cmap.N)
            
            # Apply to edges
            edge_colors = [base_cmap(norm(val)) for val in color_values]

        else:
            
            norm = Normalize(vmin=0, vmax=max(values))
            base_cmap = LinearSegmentedColormap.from_list("WhiteToBlue", [(1, 1, 1), (0, 0, 1)])
            
        color_values = [G.edges[edge][color] for edge in G.edges]

    edge_colors = [base_cmap(norm(val)) for val in color_values]

    # Create a figure and main axis; add a colorbar to the same figure.
    fig, ax = plt.subplots(figsize=(6,6))

    # Plot arcs, edges, and nodes
    for node in G.nodes:
        children = list(G.successors(node))
        if len(children) > 1:
            parent_angle = pos_polar[node][1]
            child_angles = [(child, abs(pos_polar[child][1] - parent_angle)) for child in children]
            furthest_child = max(child_angles, key=lambda x: x[1])[0]
            r0, theta0 = pos_polar[node]
            r1, theta1 = pos_polar[furthest_child]
            arc_radius = r1 - 1.0
            arc_theta = np.linspace(theta0, theta1, 10)
            arc_x = arc_radius * np.cos(arc_theta)
            arc_y = arc_radius * np.sin(arc_theta)
            ax.plot(arc_x, arc_y, color="black", linewidth=1.5, zorder=3)

    for edge_idx, edge in enumerate(G.edges):
        start, end = edge
        r0, theta0 = pos_polar[start]
        r1, theta1 = pos_polar[end]
        x_mid2, y_mid2 = polar_to_cartesian(r1 - 1.0, theta1)
        x1, y1 = polar_to_cartesian(r1, theta1)
        ax.plot([x_mid2, x1], [y_mid2, y1], color='black', linewidth=1.5, zorder=4)
        ax.scatter(x1, y1, s=20, c="k", zorder=5)
        ax.scatter(x1, y1, s=5, c=[edge_colors[edge_idx]], zorder=6)

    ax.scatter(0, 0, s=25, c="k", zorder=10, alpha=0.5)

    for radius in range(1, max_generation + radius_offset + 1):
        circle = plt.Circle((0, 0), radius, color='grey', linewidth=1.0, fill=False, 
                            linestyle=(0, (1, 5)), zorder=-3)
        ax.add_artist(circle)
        if generation_labels and radius in generation_labels:
            label_text = f"generation {radius}" if radius == generation_labels[-1] else f"{radius}"
            char_spacing = 0.3
            start_angle = start_angle_label
            label_length = (len(label_text) - 1) * char_spacing
            start_arc_length = (math.pi * radius * (start_angle / 360)) - (label_length / 2)
            for i, char in enumerate(label_text):
                arc_length = start_arc_length + i * char_spacing
                angle = 180 - (arc_length / (math.pi * radius) * 360)
                x = radius * math.cos(math.radians(angle))
                y = radius * math.sin(math.radians(angle))
                ax.text(x, y, " ", ha="center", va="center", rotation=angle - 90,
                        rotation_mode="anchor", fontsize=8, zorder=-2,
                        bbox=dict(boxstyle="round,pad=0.5", edgecolor="none", facecolor="white"))
                ax.text(x, y, char, ha="center", va="center", rotation=angle - 90,
                        rotation_mode="anchor", fontsize=8, color="grey", zorder=-1)

    ax.set_xlim(-(max_generation + radius_offset + 0.5), max_generation + radius_offset + 0.5)
    ax.set_ylim(-(max_generation + radius_offset + 0.5), max_generation + radius_offset + 0.5)
    ax.axis("off")

    # Create a ScalarMappable for the colorbar using the same normalization and cmap
    if color == "hamming_distance":
        sm = plt.cm.ScalarMappable(cmap=base_cmap, norm=plt.Normalize(vmin=0, vmax=max_color))
    else:
        sm = plt.cm.ScalarMappable(cmap=base_cmap, norm=norm)
    sm.set_array([])

    cbar = fig.colorbar(sm, ax=ax, fraction=0.046, pad=0.04)
    if color == "combined_score":
        cbar.set_label('combined score')
    if color == "final_score":
        cbar.set_label('final score')
    elif color == "BioDC_redox":
        cbar.set_label('Difference to target potential')
        cbar.ax.yaxis.set_major_formatter(plt.FuncFormatter(lambda x, _: f'{x:.2f}'))
        """
        tick_vals = [
            vmin,
            vmin + 0.5 * (vmid - vmin),
            vmid,
            vmid + 0.5 * (vmax - vmid),
            vmax
        ]
        cbar.set_ticks(tick_vals)
        cbar.set_ticklabels([f"{tv:.2f}" for tv in tick_vals])
        """
        
    elif color == "hamming_distance":
        cbar.set_label('Hamming Distance')
    else:
        cbar.set_label(color)
        
    # Always save to self.FOLDER_HOME with the naming scheme
    save_path = os.path.join(self.FOLDER_HOME, "plots", f"plot_tree_{color}.png")
    fig.savefig(save_path, bbox_inches='tight', dpi=300)


"""
Prepares commands to run protein structure prediction using AlphaFold3.

Functions:
    seq_to_json():              Converts a sequence file into a JSON input for AlphaFold3.
    prepare_AlphaFold3_MSA():   Prepares the AlphaFold3 MSA stage.
    prepare_AlphaFold3_INF():   Prepares the AlphaFold3 inference stage.
"""

import os
import time

from helper_002 import generate_remark_from_all_scores_df

def seq_to_json(self, seq_input, working_dir):
    
    """
    Converts a sequence file (with a .seq extension) into a JSON input file for AlphaFold3.

    Parameters:
        seq_input (str):    Input sequence file location
        working_dir (str):  working dir for AlphaFold3
    """

    for attempt in range(10): 
        try:
            with open(f'{seq_input}.seq', "r") as f:
                sequence = f.read()
            break  
        except IOError:
            time.sleep(1)
    else:
        raise IOError(f"Failed to read {file_path} after 10 attempts.")

    # Checks if AlphaFold3MSA is run with or without MSA
    if 'AlphaFold3MSA' in self.DESIGN_METHODS:
        MSA = ""
    else:
        MSA = """,
        "unpairedMsa": "",
        "pairedMsa": "",
        "templates": []
"""
      
    json = f'''
    {{
  "name": "{self.WT}",
  "sequences": [
    {{
      "protein": {{
        "id": ["A"],
        "sequence": "{sequence}"{MSA}
      }}
    }},
    {{
      "ligand": {{
        "id": ["X"],
        "ccdCodes": ["{self.LIGAND}"]
      }}
    }}
  ],
  "modelSeeds": [1],
  "dialect": "alphafold3",
  "version": 1
}}
'''
    
    with open(f'{working_dir}.json', "w") as f: 
        f.write(json)

def prepare_AlphaFold3_MSA(self, 
                           index, 
                           cmd):
    """
    Performs MSA for structure prediction with AlphaFold3.
    
    Parameters:
        index (str):    The index of the protein variant to be predicted.
        cmd (str):      Growing list of commands to be exected by run_design using submit_job.

    Returns:
        cmd (str):      Command to be exected by run_design using submit_job.
    """
    
    working_dir = f'{self.FOLDER_DESIGN}/{index}'

    # Find input sequence !!!! MIGHT NOT BE CORRECT !!!
    PDB_input = self.all_scores_df.at[int(index), "step_input_variant"]
    seq_input = f"{working_dir}/{self.WT}_{index}"
    
    # Make directories
    os.makedirs(f"{working_dir}/scripts", exist_ok=True)
    os.makedirs(f"{working_dir}/AlphaFold3", exist_ok=True)

    # Update working dir
    working_dir = f"{working_dir}/AlphaFold3"

    # Create json from sequence
    seq_to_json(self, seq_input, f"{working_dir}/{self.WT}_{index}")
    
    cmd += f"""### AlphaFold3_MSA ###

# Modules to load
module purge
module load apptainer/1.3.2
module load alphafold/3.0.1

# AlphaFold3 specific variables
AF3_MODEL_DIR={self.FOLDER_Alphafold}
AF3_JAX_CACHE_DIR={working_dir}/jaxcache

# Model specific variables
AF3_MSA_OUTPUT_DIR={working_dir}/MSA
AF3_MSA_JSON_PATH={working_dir}/{self.WT}_{index}.json

mkdir -p $AF3_JAX_CACHE_DIR $AF3_INFERENCE_OUTPUT_DIR

export OMP_NUM_THREADS=1

apptainer --quiet exec --bind /u:/u,/ptmp:/ptmp,/raven:/raven \
    ${{AF3_IMAGE_SIF}} \
    python3 /app/alphafold/run_alphafold.py \
    --db_dir $AF3_DB_DIR \
    --model_dir $AF3_MODEL_DIR \
    --json_path $AF3_MSA_JSON_PATH \
    --output_dir $AF3_MSA_OUTPUT_DIR \
    --norun_inference
"""
    return cmd
    
def prepare_AlphaFold3_INF(self, 
                           index, 
                           cmd,
                           gpu_id = None):
    """
    Performs MSA for structure prediction witt AlphaFold3.
    
    Parameters:
        index (str):    The index of the protein variant to be predicted.
        cmd (str):      Growing list of commands to be exected by run_design using submit_job.

    Returns:
        cmd (str):      Command to be exected by run_design using submit_job.
    """
    working_dir = f'{self.FOLDER_DESIGN}/{index}'

    # Find input sequence !!!! MIGHT NOT BE CORRECT !!!
    PDB_input = self.all_scores_df.at[int(index), "step_input_variant"]
    seq_input = f"{working_dir}/{self.WT}_{index}"
            
    # Make directories
    os.makedirs(f"{working_dir}/scripts", exist_ok=True)
    os.makedirs(f"{working_dir}/AlphaFold3", exist_ok=True)

    # Update working dir
    working_dir = f"{working_dir}/AlphaFold3"

    # Create json from sequence
    seq_to_json(self, seq_input, f"{working_dir}/{self.WT}_{index}")
    
    cmd += f"""### AlphaFold3_MSA ###
"""
    # Set GPU
    if  gpu_id != None:
        cmd += f"""
export CUDA_VISIBLE_DEVICES={gpu_id}
"""

    # Checks if AlphaFold3MSA is run with or without MSA
    if 'AlphaFold3MSA' in self.DESIGN_METHODS:
        MSA = f"{working_dir}/MSA/{self.WT}/{self.WT}_data.json"
    else:
        MSA = f"{working_dir}/{self.WT}_{index}.json"
       
    cmd += f"""
# Modules to load
module purge
module load apptainer/1.3.2
module load alphafold/3.0.1

# AlphaFold3 specific variables
AF3_MODEL_DIR={self.FOLDER_Alphafold}
AF3_JAX_CACHE_DIR={working_dir}/jaxcache

# Model specific variables  
AF3_INFERENCE_OUTPUT_DIR={working_dir}/INF
AF3_INFERENCE_JSON_PATH={MSA}

mkdir -p $AF3_JAX_CACHE_DIR $AF3_INFERENCE_OUTPUT_DIR

export OMP_NUM_THREADS=1
export XLA_FLAGS="--xla_gpu_enable_triton_gemm=false"
export XLA_PYTHON_CLIENT_PREALLOCATE=false
export TF_FORCE_UNIFIED_MEMORY=true
export XLA_CLIENT_MEM_FRACTION="3.0"

apptainer --quiet exec --bind /u:/u,/ptmp:/ptmp,/raven:/raven --nv \
    ${{AF3_IMAGE_SIF}} \
    python3 /app/alphafold/run_alphafold.py \
    --db_dir $AF3_DB_DIR \
    --model_dir $AF3_MODEL_DIR \
    --json_path $AF3_INFERENCE_JSON_PATH \
    --output_dir $AF3_INFERENCE_OUTPUT_DIR \
    --jax_compilation_cache_dir $AF3_JAX_CACHE_DIR \
    --norun_data_pipeline

{self.bash_args}python {self.FOLDER_PARENT}/cif_to_pdb.py \
--cif_file {working_dir}/INF/{self.WT.lower()}/{self.WT.lower()}_model.cif \
--pdb_file {working_dir}/{self.WT}_AlphaFold3INF_{index}.pdb

sed -i -e '/        H  /d' \
       -e '/SEQRES/d' \
       -e '/HEADER/d' {working_dir}/../{self.WT}_AlphaFold3INF_{index}.pdb       
"""

    if self.CST_NAME is not None:
        remark = generate_remark_from_all_scores_df(self, index)
        cmd += f"""
echo '{remark}' > {self.WT}_AlphaFold3INF_{index}.pdb"""

    if self.LIGAND == 'HEM':
        cmd += f"""
echo 'HETNAM     HEM X   1  HEM  ' >> {self.WT}_AlphaFold3INF_{index}.pdb"""

    cmd += f"""
cat {working_dir}/{self.WT}_AlphaFold3INF_{index}.pdb >> {self.WT}_AlphaFold3INF_{index}.pdb

sed -i -e 's/^\(ATOM.\{{17\}}\) /\\1A/'\
       -e '/        H  /d' -e '/TER/d'\
       -e '/ HEM /s/^HETATM/ATOM  /'\
       -e 's/{self.LIGAND} A/{self.LIGAND} X/g'\
       -e 's/HIE/HIS/g' \
       -e 's/HID/HIS/g' {self.WT}_MDMin_{index}.pdb
"""
    
    if self.REMOVE_TMP_FILES:
        cmd += f"""# Removing temporary directory
rm -r AlphaFold3
"""

    return cmd   


import os
import numpy as np
from Bio.PDB import PDBParser
import ipywidgets as widgets
from IPython.display import display
from bqplot import Figure, Scatter, LinearScale, Axis
import py3Dmol
from IPython.display import clear_output

from helper_002 import normalize_scores

def make_structure_viewer(self):
    out = widgets.Output(layout=widgets.Layout(height='300px', flex='1 1 auto', overflow='hidden'))
    with out:
        pdb = self.all_scores_df.loc[self.index, "final_variant"]
        pdb_file = os.path.abspath(f"{pdb}.pdb")
        display(widgets.HTML(f"<span style='font-size:12px;color:#666;'>{os.path.basename(pdb_file)}</span>"))
        if not os.path.isfile(pdb_file):
            raise FileNotFoundError(f"{pdb_file} not found")

        # Compute principal axis
        parser = PDBParser(QUIET=True)
        struct = parser.get_structure("X", pdb_file)
        coords = np.array([a.get_coord() for a in struct.get_atoms()])
        centroid = coords.mean(axis=0)
        cov = np.cov(coords - centroid, rowvar=False)
        evals, evecs = np.linalg.eigh(cov)
        principal = evecs[:, np.argmax(evals)]

        # Rotation to x-axis
        x_axis = np.array([1,0,0])
        v = np.cross(principal, x_axis)
        s = np.linalg.norm(v); c = np.dot(principal, x_axis)
        if s < 1e-6:
            angle_deg, axis_obj = 0.0, {'x':1,'y':0,'z':0}
        else:
            axis = v/s
            angle_deg = np.degrees(np.arccos(np.clip(c, -1, 1)))
            axis_obj = {'x':axis[0],'y':axis[1],'z':axis[2]}

        # 3Dmol render
        view = py3Dmol.view(width=600, height=300)
        view.addModel(open(pdb_file).read(), "pdb")
        view.setStyle({}, {"cartoon":{"color":"white"}})
        view.translate(-centroid[0], -centroid[1], -centroid[2])
        view.rotate(angle_deg, axis_obj)
        view.translate(centroid[0], centroid[1], centroid[2])

        # Highlight ligand/design
        backbone = ["C","N","O"]
        if getattr(self, 'LIGAND', None):
            view.addStyle(
                {"resn":self.LIGAND, "not":{"atom":backbone}},
                {"stick":{"radius":0.5}}
            )
        design = [r.strip() for r in getattr(self,'DESIGN',"").split(",") if r.strip()]
        if design:
            view.addStyle(
                {"resi":design, "not":{"atom":backbone}},
                {"stick":{"radius":0.3}}
            )
        rest_sel = {"and":[
            {"not":{"atom":backbone}},
            {"not":{"resi":design}},
            {"not":{"resn":getattr(self,'LIGAND',"")}}
        ]}
        view.addStyle(rest_sel, {"stick":{"radius":0.1}})
        view.zoomTo()
        view.show()
    return out

def make_score_plot(self):
    df = self.plot_scores_df
    x = np.arange(len(df))
    y = df[self.SCORE].values

    x_sc, y_sc = LinearScale(), LinearScale()

    # 1) grey base
    base = Scatter(x=x, y=y, scales={'x':x_sc,'y':y_sc},
                   default_size=32, colors=['#888888'], tooltip=widgets.Label())

    # 2) blue selected
    if self.index in df.index:
        i = df.index.get_loc(self.index)
        sel_x = [i]
        sel_y = [df.loc[self.index, self.SCORE]]
    else:
        sel_x = []
        sel_y = []

    sel = Scatter(x=sel_x, y=sel_y, scales={'x':x_sc,'y':y_sc},
                  default_size=48, colors=['#0000FF'], tooltip=widgets.Label())

    # Hover on both
    def hover_cb(_, pt):
        idx = df.index[pt['index']]
        v = df.loc[idx, self.SCORE]
        txt = f"Index: {idx}\n{self.SCORE}: {v:.3g}"
        base.tooltip.value = sel.tooltip.value = txt
    base.on_hover(hover_cb)
    sel.on_hover(hover_cb)

    # Click on both
    def click_cb(_, pt):
        idx = df.index[pt['index']]
        self.index_slider.value = idx
    base.on_element_click(click_cb)
    sel.on_element_click(click_cb)

    axes = [
        Axis(scale=x_sc, visible=False),  # hide x-axis
        Axis(label=self.SCORE, scale=y_sc, orientation='vertical')
    ]

    return Figure(marks=[base, sel],
                  axes=axes,
                  title=self.SCORE,
                  layout=widgets.Layout(width='300px', height='300px'))

def display_variants(self):
    # filter and sort
    self.plot_scores_df = self.all_scores_df[self.all_scores_df[f'{self.SELECTED_SCORES[0]}_score'].notna()]

    # remove outliers beyond ±5 standard deviations
    for score in self.SELECTED_SCORES:
        score_values = self.plot_scores_df[f'{score}_score']
        mean = score_values.mean()
        std = score_values.std()
        self.plot_scores_df = self.plot_scores_df[(score_values >= mean - 5 * std) & (score_values <= mean + 5 * std)]

    # Calculate normalized scores using your normalize_scores function.
    scores = normalize_scores(self, unblocked_all_scores_df=self.plot_scores_df, norm_all=False)
    for score in scores:
        self.plot_scores_df[score] = scores[score]
        
    # Remove bad structures below 3 standard deviations
    for score in ['final']+[i for i in self.SELECTED_SCORES if i not in ['identical','catalytic']]:
        score = f'{score}_score'
        mean = self.plot_scores_df[score].mean()
        std = self.plot_scores_df[score].std()
        self.plot_scores_df = self.plot_scores_df[(self.plot_scores_df[score] >= mean - 3 * std)]

    # Sort df
    self.plot_scores_df = self.plot_scores_df.sort_values(by=self.SCORE, ascending=True)
    opts = self.plot_scores_df.index.tolist()

    # Pre-select best scoring (first in sorted list)
    if self.index == None: 
        self.index = opts[-1]

    # controls
    if not hasattr(self, 'index_slider'):
        
        self.index_slider = widgets.SelectionSlider(
            options=opts, value=self.index,
            description="Index:", continuous_update=False,
            layout=widgets.Layout(flex='1 1 auto')
        )
        first_btn = widgets.Button(description='⏮ First')
        prev_btn = widgets.Button(description='◀ Prev')
        next_btn = widgets.Button(description='Next ▶')
        last_btn = widgets.Button(description='Last ⏭') 

        def on_first(_):
            if opts:
                self.index_slider.value = opts[0]
        
        def on_last(_):
            if opts:
                self.index_slider.value = opts[-1]
        
        def on_prev(_):
            if self.index in opts:
                i = opts.index(self.index)
            else:
                i = 0
            if i > 0:
                self.index_slider.value = opts[i - 1]
        
        def on_next(_):
            if self.index in opts:
                i = opts.index(self.index)
            else:
                i = 0
            if i < len(opts) - 1:
                self.index_slider.value = opts[i + 1]

        first_btn.on_click(on_first)
        last_btn.on_click(on_last)
        prev_btn.on_click(on_prev)
        next_btn.on_click(on_next)
        
        control = widgets.HBox(
            [self.index_slider, first_btn, prev_btn, next_btn, last_btn],
            layout=widgets.Layout(width='100%')
        )

        def slider_cb(change):
            if change['name'] == 'value' and change['new'] != self.index:
                self.index = change['new']
                self.hbox.children = (
                    make_structure_viewer(self),
                    make_score_plot(self)
                )

        self.index_slider.observe(slider_cb)
        self.controls = control
    else:
        self.index_slider.options = opts
        self.index_slider.value = self.index

    # panels
    struct = make_structure_viewer(self)
    plot = make_score_plot(self)

    if not hasattr(self, 'hbox'):
        self.hbox = widgets.HBox([struct, plot],
            layout=widgets.Layout(
                height='300px', width='100%',
                display='flex', flex_flow='row nowrap',
                align_items='flex-start'
            ))
    else:
        self.hbox.children = (struct, plot)

    # main
    clear_output(wait=True) 
    self.container = widgets.VBox([self.controls, self.hbox],
                                  layout=widgets.Layout(width='100%'))
    display(self.container)



import os
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import curve_fit
from scipy.stats import pearsonr
import networkx as nx
from sklearn.decomposition import PCA
from networkx.drawing.nx_agraph import graphviz_layout
from helper_001               import *

#import umap
#import torch
#import esm

#Imports ALL_SCORES_CSV as the plot_scores_df dataframe
def load_plot_scores_df(self):
   
    # Condition to check if the ALL_SCORES_CSV file exists, otherwise it returns the function.
    if not os.path.isfile(f'{self.FOLDER_HOME}/all_scores.csv'): 
        print(f"ERROR: {self.FOLDER_HOME}/all_scores.csv does not exist!")
        return    
    
    self.plot_scores_df = pd.read_csv(self.ALL_SCORES_CSV)
    self.plot_scores_df = self.plot_scores_df.dropna(subset=['total_score'])
    self.plot_scores_df['sequence'] =  self.plot_scores_df['sequence'].astype(str)
    self.plot_scores_df['design_method'] =  self.plot_scores_df['design_method'].astype(str)
    self.plot_scores_df['score_taken_from'] =  self.plot_scores_df['score_taken_from'].astype(str) 
    
#Define the hamming distance function and other required functions
def hamming_distance(seq1, seq2):
    #Ensures that seq2 is a string
    if not isinstance(seq2, str):
        return None
     #Ensures that the current and predecessor sequence length is equal
    if len(seq1) != len(seq2):
        raise ValueError("Sequences must be of equal length")
    #Returns the number of differences between the current sequence and the parent sequence.
    return sum(ch1 != ch2 for ch1, ch2 in zip(seq1, seq2))

def exponential_func(x, A, k, c):
    return c-A*np.exp(-k * x)


#Define all the required plots for the main plot figure output
def plot_combined_score(self, ax, combined_scores, combined_score_min, combined_score_max, combined_score_bin):
    
    ax.hist(combined_scores, bins=np.arange(combined_score_min,combined_score_max+combined_score_bin,combined_score_bin))
    
    #Add the two values HIGHSCORE and NEG_BEST as reference values for the top ever design and the best ever negative control value.
    ax.axvline(self.HIGHSCORE_NEGBEST['HIGHSCORE_combined_score'], color='b', label='Highest Score')
    ax.axvline(self.HIGHSCORE_NEGBEST['NEGBEST_combined_score'], color='r', label='Negative Best')
    
    #Sets plot details
    ax.set_xlim(combined_score_min,combined_score_max)
    ax.set_title('Histogram of Combined Score')
    ax.set_xlabel('Combined Score')
    ax.set_ylabel('Frequency')

def plot_interface_score(ax, interface_scores, interface_score_min, interface_score_max, interface_score_bin):
    ax.hist(interface_scores, density=True,
            bins=np.arange(interface_score_min,interface_score_max+interface_score_bin,interface_score_bin))
    ax.set_xlim(interface_score_min,interface_score_max)
    ax.set_title('Histogram of Interface Score')
    ax.set_xlabel('Interface Score')
    ax.set_ylabel('Frequency')

def plot_total_score(ax, total_scores, total_score_min, total_score_max, total_score_bin):
    ax.hist(total_scores, density=True,
            bins=np.arange(total_score_min,total_score_max+total_score_bin,total_score_bin))
    ax.set_xlim(total_score_min,total_score_max)
    ax.set_title('Histogram of Total Score')
    ax.set_xlabel('Total Score')
    ax.set_ylabel('Frequency')

def plot_catalytic_score(ax, catalytic_scores, total_score_min, total_score_max, total_score_bin):
    ax.hist(catalytic_scores, density=True, bins=np.arange(total_score_min,total_score_max+total_score_bin,total_score_bin))
    ax.set_xlim(total_score_min,total_score_max)
    ax.set_title('Histogram of Catalytic Score')
    ax.set_xlabel('Catalytic Score')
    ax.set_ylabel('Frequency')
    
def plot_efield_score(ax, efield_scores, total_score_min, total_score_max, total_score_bin):
    ax.hist(efield_scores, density=True, bins=np.arange(total_score_min,total_score_max+total_score_bin,total_score_bin))
    ax.set_xlim(total_score_min,total_score_max)
    ax.set_title('Histogram of Efield Score')
    ax.set_xlabel('Efield Score')
    ax.set_ylabel('Frequency')
    
def plot_boltzmann_histogram(self, ax, combined_scores, score_min, score_max, score_bin):
    
    # Generation of the combined potentials values
    _, _, _, _, combined_potentials = normalize_scores(self,
                                                       self.plot_scores_df, 
                                                       print_norm=False,
                                                       norm_all=False,
                                                       extension="score")
    
    # Definition of the kbt value which is used as the Boltzmann weight. 
    # The value is not constant throughout the simulation but decays with time (increasing index).
    # The dacay rate is defined in the second line of the code.
    if isinstance(self.KBT_BOLTZMANN, (float, int)):
        kbt_boltzmann = self.KBT_BOLTZMANN
    else:
        if len(self.KBT_BOLTZMANN) == 2:
            kbt_boltzmann = max(self.KBT_BOLTZMANN[0] * np.exp(-self.KBT_BOLTZMANN[1]*self.plot_scores_df['index'].max()), 0.05)
    
    #Calculates the weights using the Boltzmann formula which are then going to be used for the generation of the boltzmann_scores
    boltzmann_factors = np.exp(combined_potentials / (kbt_boltzmann)) 
    print(f"Min/Max boltzmann factors: {min(boltzmann_factors)}, {max(boltzmann_factors)}")
    probabilities = boltzmann_factors / sum(boltzmann_factors) 
    
    #Lists of values sampled from the combined potential either randomly (random_scores) or using weights (boltzmann_scores)
    random_scores = np.random.choice(combined_potentials, size=10000, replace=True)
    boltzmann_scores = np.random.choice(combined_potentials, size=10000, replace=True, p=probabilities)

    #Definition of the first plot using random sampling for the Boltzmann distribution
    ax.hist(random_scores, density=True, alpha=0.7, label='Random Sampling', \
            bins=np.arange(score_min-2,score_max+1+score_bin,score_bin))
    ax.text(0.05, 0.95, "normalized only to \n this dataset")
    ax.set_xlabel('Potential')
    ax.set_ylabel('Density (Normal)')
    ax.set_title(f'kbT = {kbt_boltzmann:.1e}')
    
    #Definition of a twin axis for the second histogram using the Boltzmann sampling with the weights for the Boltzmann distribution
    ax_dup = ax.twinx()
    ax_dup.hist(boltzmann_scores, density=True, alpha=0.7, color='orange', label='Boltzmann Sampling', \
                bins=np.arange(score_min-2,score_max+1+score_bin,score_bin))
    ax.set_xlim(score_min-2,score_max+1)
    ax_dup.set_ylabel('Density (Boltzmann)')
    ax_dup.tick_params(axis='y', labelcolor='orange')
    
def plot_combined_score_v_index(self, ax, combined_scores):
    
    #Creates a scatter plot of the combined scores versus the index.
    combined_scores = pd.Series(combined_scores)
    ax.scatter(self.plot_scores_df['index'], combined_scores, c='lightgrey', s=5) 
    
    #Adds the two values HIGHSCORE and NEG_BEST as reference values for the top ever design and the best ever negative control value.
    ax.axhline(self.HIGHSCORE_NEGBEST['HIGHSCORE_combined_score'], color='b', label='Highest Score', alpha = 0.5)
    ax.axhline(self.HIGHSCORE_NEGBEST['NEGBEST_combined_score'], color='r', label='Negative Best', alpha = 0.5)
    
    #Imports values from the PARENT files and creates a x-line on the plot indicating ????
    full_path = os.path.join(self.FOLDER_HOME, self.FOLDER_PARENT)
    PARENTS = [i for i in os.listdir(full_path) if i.endswith(".pdb") and os.path.isfile(os.path.join(full_path, i))]
    ax.axvline(self.N_PARENT_JOBS*len(PARENTS), color='k')
    
    #Calculates the moving average in a window of 20 of the combined scores and creates a line plot with the moving average values.
    moving_avg = combined_scores.rolling(window=20).mean()
    ax.plot(range(len(moving_avg)),moving_avg,c="k")
    
    #Sets plot details
    ax.set_ylim(0,1)
    ax.set_xlim(0,self.MAX_DESIGNS)
    ax.set_title('Combined score vs Index')
    ax.set_xlabel('Index')
    ax.set_ylabel('Combined Score')
    ax.legend(loc='best')
    
def plot_combined_score_v_generation_violin(self, ax, combined_scores):
    
    #Defines the values for generating the violin plot and creates it
    self.plot_scores_df['tmp'] = combined_scores
    max_gen = int(self.plot_scores_df['generation'].max())
    generations = np.arange(0, max_gen + 1)
    violin_data = [self.plot_scores_df[self.plot_scores_df['generation'] == gen]['tmp'].dropna().values for gen in generations]
    parts = ax.violinplot(violin_data, positions=generations, showmeans=False, showmedians=True)
    
    # Customizing the color of violin plots
    for pc in parts['bodies']:
        pc.set_facecolor('green')
        pc.set_edgecolor('black')
        pc.set_alpha(0.7)
    
    # Customizing the color of the median lines
    for partname in ('cbars', 'cmins', 'cmaxes'):
        vp = parts.get(partname)
        if vp:
            vp.set_edgecolor('tomato')
            vp.set_linewidth(0.5)
    vp = parts.get('cmedians')
    if vp:
        vp.set_edgecolor('tomato')
        vp.set_linewidth(2.0)
    
    # Fit the data to the exponential function and plot it
    #weights = np.linspace(1, 0.1, len(generations))
    #weights = np.ones(len(generations))
    #weights[:1] = 0.3
    #mean_scores = [np.mean(data) for data in violin_data]
    #popt, pcov = curve_fit(exponential_func, generations, mean_scores, p0=(1, 0.1, 0.7), sigma=weights, maxfev=2000)
    #fitted_curve = exponential_func(generations, *popt)
    #ax.plot(generations, fitted_curve, 'r--', label=f'Fit: A*exp(-kt) - c\nA={popt[0]:.2f}, k={popt[1]:.2f}, c={popt[2]:.2f}')
    
    #Adds the two values HIGHSCORE and NEG_BEST as reference values for the top ever design and the best ever negative control value.
    ax.axhline(self.HIGHSCORE_NEGBEST['HIGHSCORE_combined_score'], color='b', label='Highest Score', alpha = 0.5)
    ax.axhline(self.HIGHSCORE_NEGBEST['NEGBEST_combined_score'], color='r', label='Negative Best', alpha = 0.5)

    #Sets plot details
    ax.set_ylim(0, 1)
    ax.set_title('Combined Score vs Generations')
    ax.set_xlabel('Generation')
    ax.set_ylabel('Combined Score')
    ax.yaxis.grid(True, linestyle='--', which='major', color='grey', alpha=0.7)
    ax.legend(loc='best')
    every_second_generation = generations[::2]
    ax.set_xticks(every_second_generation)
    ax.set_xticklabels(every_second_generation)
    
def plot_score_v_generation_violin(self, ax, score_type):
    
    #Prepares data for the violin plot generation and creates it
    max_gen = int(self.plot_scores_df['generation'].max())
    generations = np.arange(0, max_gen + 1)
    violin_data = [self.plot_scores_df[self.plot_scores_df['generation'] == gen][score_type].dropna().values for gen in generations]
    parts = ax.violinplot(violin_data, positions=generations, showmeans=False, showmedians=True)
    
    # Customizing the color of violin plots
    for pc in parts['bodies']:
        pc.set_facecolor('green')
        pc.set_edgecolor('black')
        pc.set_alpha(0.7)
        
    # Customizing the color of the median lines
    for partname in ('cbars', 'cmins', 'cmaxes'):
        vp = parts.get(partname)
        if vp:
            vp.set_edgecolor('tomato')
            vp.set_linewidth(0.5)
    vp = parts.get('cmedians')
    if vp:
        vp.set_edgecolor('tomato')
        vp.set_linewidth(2.0)
 
    # Fit the data to the exponential function and plot it
    # weights = np.ones(len(generations))
    # mean_scores = [np.mean(data) for data in violin_data]
    # popt, pcov = curve_fit(exponential_func, generations, mean_scores, p0=(1, 0.1, 0.7), sigma=weights, maxfev=10000)
    # fitted_curve = exponential_func(generations, *popt)
    # ax.plot(generations, fitted_curve, 'r--', label=f'Fit: A*exp(-kt) - c\nA={popt[0]:.2f}, k={popt[1]:.2f}, c={popt[2]:.2f}')
    
    #Adds the two values HIGHSCORE and NEG_BEST as reference values for the top ever design and the best ever negative control value.
    ax.axhline(self.HIGHSCORE_NEGBEST[f'HIGHSCORE_{score_type}'], color='b', label='Highest Score', alpha=0.5)
    ax.axhline(self.HIGHSCORE_NEGBEST[f'NEGBEST_{score_type}'], color='r', label='Negative Best', alpha=0.5)

    #Sets the plot details
    ax.set_title(f'{score_type.replace("_", " ").title()} vs Generations')
    ax.set_xlabel('Generation')
    ax.set_ylabel(f'{score_type.replace("_", " ").title()}')
    ax.yaxis.grid(True, linestyle='--', which='major', color='grey', alpha=0.7)
    ax.legend(loc='best')
    every_fourth_generation = generations[::4]
    ax.set_xticks(every_fourth_generation)
    ax.set_xticklabels(every_fourth_generation)
    
def plot_mutations_v_generation_violin(self, ax, mut_min, mut_max):
    
    #Prepare data for the violin plot generation and creates it
    max_gen = int(self.plot_scores_df['generation'].max())
    generations = np.arange(0, max_gen + 1)
    violin_data = [self.plot_scores_df[self.plot_scores_df['generation'] == gen]['mutations'].dropna().values for gen in generations]
    parts = ax.violinplot(violin_data, positions=generations, showmeans=False, showmedians=True)
    ax.axhline(len(self.DESIGN.split(",")), color='r', label='Max # Mutations')
    
    # Customizing the color of violin plots
    for pc in parts['bodies']:
        pc.set_facecolor('green')
        pc.set_edgecolor('black')
        pc.set_alpha(0.7)
        
    # Customizing the color of the median lines
    for partname in ('cbars', 'cmins', 'cmaxes'):
        vp = parts.get(partname)
        if vp:
            vp.set_edgecolor('tomato')
            vp.set_linewidth(0.5)
    vp = parts.get('cmedians')
    if vp:
        vp.set_edgecolor('tomato')
        vp.set_linewidth(2.0)

    # Fit the data to the exponential function and plot the fitted curve
    # mean_mutations = [np.mean(data) for data in violin_data]
    # weights = np.ones(len(generations))  # Uniform weights, adjust as needed
    # popt, pcov = curve_fit(exponential_func, generations, mean_mutations, p0=(1, 0.1, 0.7), sigma=weights, maxfev=2000)
    # fitted_curve = exponential_func(generations, *popt)
    # ax.plot(generations, fitted_curve, 'r--', label=f'Fit: A*exp(-kt) + c\nA={popt[0]:.2f}, k={popt[1]:.2f}, c={popt[2]:.2f}')
    
    #Sets plot details
    ax.set_ylim(mut_min, mut_max)
    ax.set_title('Mutations vs Generations')
    ax.set_xlabel('Generation')
    ax.set_ylabel('Number of Mutations')
    ax.yaxis.grid(True, linestyle='--', which='major', color='grey', alpha=0.7)
    ax.legend(loc='lower right')
    every_second_generation = generations[::2]
    ax.set_xticks(every_second_generation)
    ax.set_xticklabels(every_second_generation)
    
def plot_hamming_distance_v_generation_violin(self, ax):
    
    #Retrieves the amino acid sequence of the ancestor protein
    with open(f"{self.FOLDER_HOME}/parent/{self.WT}.seq", "r") as f:
            ancestor_sequence = f.read()

    #Defines the data relative to the generations
    max_gen = int(self.plot_scores_df['generation'].max())
    generations = np.arange(0, max_gen + 1)
    
    #Creates an empty column in the plot_scores_df dataframe where the hamming distances will be uploaded
    self.plot_scores_df['Hamming Distance'] = 0  

    # Iterates through the rows and calculates the hamming distances values
    for index, row in self.plot_scores_df.iterrows():
        index = int(float(row['index'])) + 1
        if not isinstance(row['sequence'], str):
            continue
            
        #If the design belongs to the generation 0, the hamming distance is calculated using the original protein sequence
        if row['parent_index'] == "Parent":
            parent_sequence = ancestor_sequence
            
        #If the design belongs to the generation 1 or higher, the hamming distance is calculated using the corresponding parent sequence
        else:
            parent_idx = int(float(row['parent_index'])) + 1
            parent_sequence = self.plot_scores_df.loc[self.plot_scores_df.index == parent_idx - 1, 'sequence'].values[0]
        current_sequence = row['sequence']
        
        #Adds the calculated hamming distance values into the last column of the dataframe, called "Hamming Distance"
        self.plot_scores_df.at[index,"Hamming Distance"] = hamming_distance(parent_sequence, current_sequence)
        
    #Creates the data used for the generation of the violin plot and generates it
    violin_data_hm = [self.plot_scores_df[self.plot_scores_df['generation'] == gen]["Hamming Distance"].dropna().values for gen in generations]
    ax.violinplot(violin_data_hm,positions=generations, showmeans=True, showmedians=False)
    
    #Sets plot details
    ax.set_title('Hamming Distance vs Generations')
    ax.set_xlabel('Generation')
    ax.set_ylabel('Hamming Distance')
    every_fourth_generation = generations[::4]
    ax.set_xticks(every_fourth_generation)
    ax.set_xticklabels(every_fourth_generation)
    
#-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------    
    
#Defines the plot_scores function which generates the 12 main plots and saves them in the plot folder
def plot_scores(self, combined_score_min=0, combined_score_max=1, combined_score_bin=0.01, 
                interface_score_min=0, interface_score_max=1, interface_score_bin=0.01,
                total_score_min=0, total_score_max=1, total_score_bin=0.01,
                catalytic_score_min=0, catalytic_score_max=1, catalytic_score_bin=0.01, print_vals=True,
                ):
      
        mut_min=0
        mut_max=len(self.DESIGN.split(","))+1 
        
        #Retrieves the scores data from the ALL_SCORES_CSV file
        load_plot_scores_df(self)        
        
        # ---------------------------------------------------------------------------------------------------------------
        # Checks whether the FOLDER_PLOT exists otherwise it creates the directory where the plots will be saved.
        # DONE DURING STARTUP - CAN BE DELETED IN THE FUTURE!
        self.FOLDER_PLOT = f'{self.FOLDER_HOME}/plots'        
        os.makedirs(self.FOLDER_PLOT, exist_ok=True)
        # ---------------------------------------------------------------------------------------------------------------
        
        
        
        #Checkpoint to see whether enough designs are completed, at least 3 are needed to generate the plots.
        if len(self.plot_scores_df) < 3: 
            print(f"ERROR: plot_scores_df does not contain enough data")
            print(self.plot_scores_df)
            return

        
        
        #to integrate??
        if (self.ProteinMPNN_PROB > 0 or self.LMPNN_PROB > 0):
            #first not nan sequence from all_scores_df
            mut_max = len(plot_scores_df[plot_scores_df['sequence'] != 'nan']['sequence'].iloc[0])    

        # Defines the scores variables by running the normalize_scores function.
        catalytic_scores, total_scores, interface_scores, efield_scores, combined_scores = normalize_scores(self,
                                                                                                            self.plot_scores_df,
                                                                                                            print_norm=True,
                                                                                                            norm_all=True)
        # Print normalized values
        if print_vals:            
            print(f'interface_score: {round(min(interface_scores),3)} , {round(max(interface_scores),3)}')
            print(f'total_score: {round(min(total_scores),3)} , {round(max(total_scores),3)}')
            print(f'efield_score: {round(min(efield_scores),3)} , {round(max(efield_scores),3)}')
            print(f'combined_score: {round(min(combined_scores),3)} , {round(max(combined_scores),3)}')

        # Creates a figure with the 12 subplots and generates the different plots. Their position in the figure is defined by the axs index.
        fig, axs = plt.subplots(3, 4, figsize=(15, 9))

        plot_combined_score(self, axs[0,0], combined_scores, combined_score_min, combined_score_max, combined_score_bin)

        plot_interface_score(axs[0,1], interface_scores, interface_score_min, interface_score_max, interface_score_bin)

        plot_total_score(axs[0,2], total_scores,  total_score_min, total_score_max, total_score_bin)

        plot_catalytic_score(axs[0,3], catalytic_scores, catalytic_score_min, catalytic_score_max, catalytic_score_bin)

        plot_efield_score(axs[1,0], efield_scores, catalytic_score_min, catalytic_score_max, catalytic_score_bin)

        plot_boltzmann_histogram(self, axs[1,1], combined_scores, combined_score_min, combined_score_max, combined_score_bin)

        plot_combined_score_v_index(self, axs[1,2], combined_scores)

        plot_combined_score_v_generation_violin(self, axs[1,3], combined_scores)

        plot_hamming_distance_v_generation_violin(self, axs[2,0])

        plot_score_v_generation_violin(self, axs[2,1], 'total_score')

        plot_score_v_generation_violin(self, axs[2,2], 'interface_score')

        plot_score_v_generation_violin(self, axs[2,3], 'efield_score')

        #Additional plot types to interchange
        #plot_mutations_v_generation_violin(self, axs[2,0], plot_scores_df, mut_min, mut_max)
        #plot_interface_score_v_total_score(axs[-,-],plot_scores_df, total_score_min, total_score_max, interface_score_min, interface_score_max)

        plt.tight_layout()
        #plt.savefig(os.path.join(self.FOLDER_PLOT, 'main_plots.png'), format='png')

        plt.show()
        
#Defines the tree_plotting function which generates the tree plot and saves it in the plot folder
def tree_plotting_function():
        
    # ---------------------------------------------------------------------------------------------------------------
    # Checks whether the FOLDER_PLOT exists otherwise it creates the directory where the plots will be saved.
    # DONE DURING STARTUP - CAN BE DELETED IN THE FUTURE!
    self.FOLDER_PLOT = f'{self.FOLDER_HOME}/plots'        
    os.makedirs(self.FOLDER_PLOT, exist_ok=True)
    # ---------------------------------------------------------------------------------------------------------------
        
    #Retrieves the scores data from the ALL_SCORES_CSV file
    load_plot_scores_df(self)   
    
    #_, _, interface_potentials, _, combined_potentials = normalize_scores(None, plot_scores_df, print_norm=False, norm_all=False, extension="potential")
    plot_scores_df["interface_potential"] = interface_potentials
    
    #Creates an empty directed graph to then be filled in
    G = nx.DiGraph()
    
    #Extracts the sequence  for the parent and current design iteratively and creates nodes for each design and an edge to the parent design equal with hamming distance
    for _, row in plot_scores_df.iterrows():
        index = int(float(row['index'])) + 1
        if not isinstance(row['sequence'], str):
            continue
        G.add_node(index, sequence=row['sequence'], interface_potential=row['interface_potential'], gen=int(row['generation']) + 1)
        if row['parent_index'] != "Parent":
            parent_idx = int(float(row['parent_index'])) + 1
            parent_sequence = plot_scores_df.loc[plot_scores_df.index == parent_idx - 1, 'sequence'].values[0]
            current_sequence = row['sequence']
            #Calculates Hamming distance between the parent and the current sequence
            distance = hamming_distance(parent_sequence, current_sequence)
            #Adds an edge between the parent and the current index with the hamming distance as an attribute
            G.add_edge(parent_idx, index, hamming_distance=distance)
    
    #Converts the directed graph G into an undirected one
    G_undirected = G.to_undirected()

    #Creates a new root node
    G.add_node(0, sequence='root', interface_potential=0, gen=0)
    
    #Connects the new root node to all nodes of generation 1
    for node in G.nodes:
        if G.nodes[node]['gen'] == 1:
            current_sequence = G.nodes[node]['sequence']
            parent_sequence = plot_scores_df.loc[plot_scores_df.index == parent_idx - 1, 'sequence'].values[0]
            distance = hamming_distance(parent_sequence, current_sequence)
            G.add_edge(0, node, hamming_distance=distance)
    
    #Uses graphviz_layout to get the positions for a concentric circular layout
    pos = graphviz_layout(G, prog="twopi", args="")

    #Normalizes scores from 0 to 1, with 0 the minimum and 1 the maximum score
    scores = {node: plot_scores_df.loc[plot_scores_df['index'] == int(node)-1, 'interface_score'].values[0] for node in G.nodes if node != 0}
    min_score = min(scores.values())
    max_score = max(scores.values())
    normalized_scores = {node: (score - min_score) / (max_score - min_score) for node, score in scores.items()}

    #Colors all nodes in black and only the root node in white.
    nodes_colors = ['white' if node == 0 else 'black' for node in G.nodes]

    #Normalizes Hamming distances for edge colors
    hamming_distances = [G.edges[edge]['hamming_distance'] for edge in G.edges]
    
    #Normalizes the Hamming distances, setting the minimal to 0 and the maximal to 1
    min_hamming = min(hamming_distances)
    max_hamming = max(hamming_distances)
    normalized_hamming = [(dist - min_hamming) / (max_hamming - min_hamming) for dist in hamming_distances]

    #Plots the tree graph
    fig, ax = plt.subplots(figsize=(10, 10), dpi=300)
    nx.draw_networkx_nodes(G, pos, ax=ax, node_size=5, linewidths=0.1, node_color=nodes_colors)
    
    #Draws edges with custom color based on normalized Hamming distance, sets color to black if hamming distance=0
    edge_colors = ['black' if norm_dist == 0 else plt.cm.cool(norm_dist) for norm_dist in normalized_hamming]
    nx.draw_networkx_edges(G, pos, ax=ax, width=0.5, edge_color=edge_colors, style='-', arrows=False)

    #Creates a colorbar as a legend for Hamming distances
    sm = plt.cm.ScalarMappable(cmap=plt.cm.cool, norm=plt.Normalize(vmin=min_hamming, vmax=max_hamming))
    sm.set_array([])
    cbar = plt.colorbar(sm, ax=ax)
    cbar.set_label('Hamming Distance')

    #Sets the plot details
    ax.set_title("Gen 0 in White, Edges colored by Hamming Distance")
    ax.axis("equal")
          
    plt.savefig(os.path.join(FOLDER_PLOT, 'tree_plot.png'), format='png')
    
    plt.show()

        
#Defines the landscape_plotting function which generates the tree plot and saves it in the plot folder
def landscape_plotting_function():
        print("landscape plot")
        
        #Checks whether the FOLDER_PLOT exists otherwise it creates the directory where the plots will be saved.
        if not os.path.exists(FOLDER_PLOT):
            os.mkdir(FOLDER_PLOT)
            
        #plt.savefig(os.path.join(FOLDER_PLOT, 'landscape_plot.png'), format='png')
        
        #plt.show()

    
    
    
    
    
    
    
    
    

       
    
    
    
    
    
    

####################################################### ##    TO WORK ON    #######################################################

def plot_summary():
    all_scores_df = pd.read_csv(ALL_SCORES_CSV)
    
    fig, axs = plt.subplots(3, 2, figsize=(20, 23))

    # Normalize scores
    catalytic_scores, total_scores, interface_scores, efield_scores, combined_scores = normalize_scores(all_scores_df, 
                                                                                         print_norm=True,
                                                                                         norm_all=True)
    
    # Plot interface vs total score colored by generation
    plot_interface_v_total_score_generation(axs[0,0], total_scores, interface_scores, all_scores_df['generation'])

    # Plot stacked histogram of interface scores by generation
    plot_stacked_histogram_by_generation(axs[0, 1], all_scores_df)

    # Plot stacked histogram of interface scores by catalytic residue index
        # Create a consistent color map for catalytic residues
    all_scores_df['cat_resi'] = pd.to_numeric(all_scores_df['cat_resi'], errors='coerce')
    unique_cat_resi = all_scores_df['cat_resi'].dropna().unique()
    colors = plt.cm.tab20(np.linspace(0, 1, len(unique_cat_resi)))
    color_map = {resi: colors[i] for i, resi in enumerate(unique_cat_resi)}
    plot_stacked_histogram_by_cat_resi(axs[1, 0], all_scores_df, color_map=color_map)

    # Plot stacked histogram of interface scores by catalytic residue index (excluding generation 0)
    all_scores_df_cleaned = all_scores_df[all_scores_df['generation'] != 0]  # Exclude generation 0
    plot_stacked_histogram_by_cat_resi(axs[1, 1], all_scores_df_cleaned, color_map=color_map)

    # Plot interface vs total score colored by catalytic residue index
    plot_interface_v_total_score_cat_resi(axs[2,0], total_scores, interface_scores, all_scores_df['cat_resi'])

    # Plot interface vs total score colored by catalytic residue name
    legend_elements = plot_interface_v_total_score_cat_resn(axs[2, 1], total_scores, interface_scores, all_scores_df['cat_resn'])
    
    axs[0,0].set_title('Total Scores vs Interface Scores by Generation')
    axs[0,1].set_title('Stacked Histogram of Interface Scores by Generation')
    axs[1,0].set_title('Stacked Histogram of Interface Scores by Catalytic Residue Index')
    axs[1,1].set_title('Stacked Histogram of Interface Scores by Catalytic Residue Index (Excluding Generation 0)')
    axs[2,0].set_title('Total Scores vs Interface Scores by Catalytic Residue Index')
    axs[2,1].set_title('Total Scores vs Interface Scores by Catalytic Residue Name')

    # Adjust legends
    # For cat_resi
    handles_cat_resi, labels_cat_resi = axs[1,1].get_legend_handles_labels()
    fig.legend(handles_cat_resi, labels_cat_resi, loc='upper right', bbox_to_anchor=(0.95, 0.65), title="Catalytic Residue Index")

    # For generation
    handles_generation, labels_generation = axs[0,1].get_legend_handles_labels()
    fig.legend(handles_generation, labels_generation, loc='upper right', bbox_to_anchor=(0.95, 0.98), title="Generation")

    # For cat_resn
    handles_cat_resn, labels_cat_resn = axs[2,1].get_legend_handles_labels()
    fig.legend(handles = legend_elements, loc='upper right', bbox_to_anchor=(0.95, 0.31), title="Catalytic Residue")

    # Adjust layout to make space for the legends on the right
    plt.tight_layout(rect=[0, 0, 0.85, 1])

def plot_interface_v_total_score_selection(ax, total_scores, interface_scores, selected_indices):
    """
    Plots a scatter plot of total_scores vs interface_scores and highlights the points
    corresponding to the selected indices.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to plot on.
    - total_scores (list or np.array): The total scores of the structures.
    - interface_scores (list or np.array): The interface scores of the structures.
    - selected_indices (list of int): Indices of the points to highlight.
    """
    
    # Create a mask for selected indices
    mask = np.ones(len(total_scores), dtype=bool)  # Initialize mask to include all points
    mask[selected_indices] = False  # Exclude selected indices
    
    # Plot all points excluding the selected ones on the given Axes object
    ax.scatter(total_scores[mask], interface_scores[mask], color='gray', alpha=0.3, label='All Points', s=1)
    
    # Highlight selected points on the given Axes object
    ax.scatter(total_scores[selected_indices], interface_scores[selected_indices], color='red', alpha=0.4, label='Selected Points', s=1)
    
    ax.set_title('Total Scores vs Interface Scores')
    ax.set_xlabel('Total Score')
    ax.set_ylabel('Interface Score')
    ax.legend()
    ax.grid(True)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

def plot_interface_v_total_score_cat_resi(ax, total_scores, interface_scores, cat_resi):
    from matplotlib.lines import Line2D
    """
    Plots a scatter plot of total_scores vs interface_scores and colors the points
    according to the catalytic residue number (cat_resi) for all data points, using categorical coloring.
    Adds a legend to represent each unique catalytic residue number with its corresponding color.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to plot on.
    - total_scores (list or np.array): The total scores of the structures.
    - interface_scores (list or np.array): The interface scores of the structures.
    - cat_resi (pd.Series or np.array): Catalytic residue numbers for all data points.
    """
    # Ensure cat_resi is a pandas Series for easier handling and remove NaN values
    if not isinstance(cat_resi, pd.Series):
        cat_resi = pd.Series(cat_resi)
    cat_resi = cat_resi.dropna()  # Drop NaN values
    
    # Proceed with the rest of the function after removing NaN values
    unique_cat_resi = cat_resi.unique()
    colors = plt.cm.tab20(np.linspace(0, 1, len(unique_cat_resi)))

    color_map = {resi: colors[i] for i, resi in enumerate(unique_cat_resi)}
    cat_resi_colors = cat_resi.map(color_map).values

    scatter = ax.scatter(total_scores[cat_resi.index], interface_scores[cat_resi.index], c=cat_resi_colors, alpha=0.4, s=2)

    legend_elements = [Line2D([0], [0], marker='o', color='w', label=f'Cat Resi {resi}',
                              markerfacecolor=color_map[resi], markersize=10) for resi in unique_cat_resi]
    #ax.legend(handles=legend_elements, title="Catalytic Residue", bbox_to_anchor=(1.05, 1), loc='upper left')

    ax.set_title('Total Scores vs Interface Scores')
    ax.set_xlabel('Total Score')
    ax.set_ylabel('Interface Score')
    ax.grid(True)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

def plot_interface_v_total_score_cat_resn(ax, total_scores, interface_scores, cat_resn):
    from matplotlib.lines import Line2D
    """
    Plots a scatter plot of total_scores vs interface_scores and colors the points
    according to the catalytic residue name (cat_resn) for all data points, using categorical coloring.
    Adds a legend to represent each unique catalytic residue name with its corresponding color.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to plot on.
    - total_scores (list or np.array): The total scores of the structures.
    - interface_scores (list or np.array): The interface scores of the structures.
    - cat_resn (pd.Series or np.array): Catalytic residue names for all data points.
    """
    # Ensure cat_resn is a pandas Series for easier handling and remove NaN values
    if not isinstance(cat_resn, pd.Series):
        cat_resn = pd.Series(cat_resn)
    cat_resn = cat_resn.dropna()  # Drop NaN values
    
    # Proceed with the rest of the function after removing NaN values
    unique_cat_resn = cat_resn.unique()
    colors = plt.cm.tab20(np.linspace(0, 1, len(unique_cat_resn)))

    color_map = {resn: colors[i] for i, resn in enumerate(unique_cat_resn)}
    cat_resn_colors = cat_resn.map(color_map).values

    scatter = ax.scatter(total_scores[cat_resn.index], interface_scores[cat_resn.index], c=cat_resn_colors, alpha=0.4, s=2)

    legend_elements = [Line2D([0], [0], marker='o', color='w', label=f'Cat Resn {resn}',
                              markerfacecolor=color_map[resn], markersize=10) for resn in unique_cat_resn]
    #ax.legend(handles=legend_elements, title="Catalytic Residue", bbox_to_anchor=(1.05, 1), loc='upper left')

    ax.set_title('Total Scores vs Interface Scores')
    ax.set_xlabel('Total Score')
    ax.set_ylabel('Interface Score')
    ax.grid(True)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    return legend_elements

def plot_interface_v_total_score_generation(ax, total_scores, interface_scores, generation):
    """
    Plots a scatter plot of total_scores vs interface_scores and colors the points
    according to the generation for all data points, using categorical coloring.
    Adds a legend to represent each unique generation with its corresponding color.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to plot on.
    - total_scores (list or np.array): The total scores of the structures.
    - interface_scores (list or np.array): The interface scores of the structures.
    - generation (pd.Series or np.array): Generation numbers for all data points.
    """
    if not isinstance(generation, pd.Series):
        generation = pd.Series(generation)
    generation = generation.dropna()  # Drop NaN values
    
    unique_generations = generation.unique()
    colors = plt.cm.viridis(np.linspace(0, 1, len(unique_generations)))

    color_map = {gen: colors[i] for i, gen in enumerate(unique_generations)}
    
    # Loop through each generation to plot, adjusting alpha for generation 0
    for gen in unique_generations:
        gen_mask = generation == gen
        alpha_value = 0.2 if gen == 0 else 0.8  # More transparent for generation 0
        ax.scatter(total_scores[generation.index][gen_mask], interface_scores[generation.index][gen_mask], 
                   c=[color_map[gen]], alpha=alpha_value, s=2, label=f'Generation {gen}' if gen == 0 else None)
    #ax.legend(handles=legend_elements, title="Generation", bbox_to_anchor=(1.05, 1), loc='upper left')

    ax.set_title('Total vs Interface Scores - Generation', fontsize=18)
    ax.set_xlabel('Total Score', fontsize=16)
    ax.set_ylabel('Interface Score', fontsize=16)
    ax.grid(True)

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)

    ax.tick_params(axis='both', which='major', labelsize=16)

def plot_stacked_histogram_by_cat_resi(ax, all_scores_df, color_map=None, show_legend=False):
    """
    Plots a stacked bar plot of interface scores colored by cat_resi on the given Axes object,
    where each bar's segments represent counts of different cat_resi values in that bin.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to plot on.
    - all_scores_df (pd.DataFrame): DataFrame containing 'cat_resi' and 'interface_score' columns.
    - color_map (dict): Optional; A dictionary mapping catalytic residue indices to colors.
    - show_legend (bool): Optional; Whether to show the legend. Defaults to False.
    """
    # Drop rows with NaN in 'interface_score'
    all_scores_df_cleaned = all_scores_df.dropna(subset=['interface_score'])

    # Ensure cat_resi is numeric and drop NaN values
    all_scores_df_cleaned['cat_resi'] = pd.to_numeric(all_scores_df_cleaned['cat_resi'], errors='coerce').dropna()
    unique_cat_resi = all_scores_df_cleaned['cat_resi'].unique()

    # Use the provided color_map or generate a new one
    if color_map is None:
        colors = plt.cm.tab20(np.linspace(0, 1, len(unique_cat_resi)))  # Generate colors for unique cat_resi values
        color_map = {resi: colors[i] for i, resi in enumerate(unique_cat_resi)}  # Map cat_resi to colors

    # Define bins for the histogram
    bins = np.linspace(all_scores_df_cleaned['interface_score'].min(), all_scores_df_cleaned['interface_score'].max(), 21)
    bin_centers = 0.5 * (bins[:-1] + bins[1:])

    # Calculate counts for each cat_resi in each bin
    counts = {resi: np.histogram(all_scores_df_cleaned[all_scores_df_cleaned['cat_resi'] == resi]['interface_score'], bins=bins)[0] for resi in unique_cat_resi}

    # Plot stacked bars for each bin
    bottom = np.zeros(len(bin_centers))
    for resi in unique_cat_resi:
        ax.bar(bin_centers, counts[resi], bottom=bottom, width=np.diff(bins), label=f'Cat Resi {resi}', color=color_map[resi], align='center')
        bottom += counts[resi]

    # Create a custom legend
    legend_elements = [Line2D([0], [0], marker='o', color='w', label=f'Cat Resi {resi}',
                              markerfacecolor=color_map[resi], markersize=10) for resi in unique_cat_resi]
    if show_legend:
        ax.legend(handles=legend_elements, title="Catalytic Residue")

    ax.set_title('Stacked Histogram of Interface Scores')
    ax.set_xlabel('Interface Score')
    ax.set_ylabel('Count')

    ax.set_xlim(-32.5, -13.5)

def plot_stacked_histogram_by_cat_resn(ax, all_scores_df):
    """
    Plots a stacked bar plot of interface scores colored by cat_resn on the given Axes object,
    where each bar's segments represent counts of different cat_resn values in that bin.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to plot on.
    - all_scores_df (pd.DataFrame): DataFrame containing 'cat_resn' and 'interface_score' columns.
    """
    # Drop rows with NaN in 'interface_score'
    all_scores_df_cleaned = all_scores_df.dropna(subset=['interface_score'])

    # Ensure cat_resn is a string and drop NaN values
    all_scores_df_cleaned['cat_resn'] = all_scores_df_cleaned['cat_resn'].astype(str).dropna()
    unique_cat_resn = all_scores_df_cleaned['cat_resn'].unique()
    colors = plt.cm.tab20(np.linspace(0, 1, len(unique_cat_resn)))  # Generate colors for unique cat_resn values

    color_map = {resn: colors[i] for i, resn in enumerate(unique_cat_resn)}  # Map cat_resn to colors

    # Define bins for the histogram
    bins = np.linspace(all_scores_df_cleaned['interface_score'].min(), all_scores_df_cleaned['interface_score'].max(), 21)
    bin_centers = 0.5 * (bins[:-1] + bins[1:])

    # Calculate counts for each cat_resn in each bin
    counts = {resn: np.histogram(all_scores_df_cleaned[all_scores_df_cleaned['cat_resn'] == resn]['interface_score'], bins=bins)[0] for resn in unique_cat_resn}

    # Plot stacked bars for each bin
    bottom = np.zeros(len(bin_centers))
    for resn in unique_cat_resn:
        ax.bar(bin_centers, counts[resn], bottom=bottom, width=np.diff(bins), label=f'Cat Resn {resn}', color=color_map[resn], align='center')
        bottom += counts[resn]
    
    ax.set_xlim(-32.5, -13.5)

    # Create a custom legend
    legend_elements = [Line2D([0], [0], marker='o', color='w', label=f'Cat Resn {resn}',
                              markerfacecolor=color_map[resn], markersize=10) for resn in unique_cat_resn]
    #ax.legend(handles=legend_elements, title="Catalytic Residue")

    ax.set_title('Stacked Histogram of Interface Scores by Catalytic Residue')
    ax.set_xlabel('Interface Score')
    ax.set_ylabel('Count')

def plot_stacked_histogram_by_generation(ax, all_scores_df):
    """
    Plots a stacked bar plot of interface scores colored by generation on the given Axes object,
    where each bar's segments represent counts of different generation values in that bin.

    Parameters:
    - ax (matplotlib.axes.Axes): The Axes object to plot on.
    - all_scores_df (pd.DataFrame): DataFrame containing 'generation' and 'interface_score' columns.
    """
    all_scores_df_cleaned = all_scores_df.dropna(subset=['interface_score', 'generation'])

    all_scores_df_cleaned['generation'] = pd.to_numeric(all_scores_df_cleaned['generation'], errors='coerce').dropna()
    unique_generations = all_scores_df_cleaned['generation'].unique()
    colors = plt.cm.viridis(np.linspace(0, 1, len(unique_generations)))

    color_map = {gen: colors[i] for i, gen in enumerate(unique_generations)}

    bins = np.linspace(all_scores_df_cleaned['interface_score'].min(), all_scores_df_cleaned['interface_score'].max(), 21)
    bin_centers = 0.5 * (bins[:-1] + bins[1:])

    counts = {gen: np.histogram(all_scores_df_cleaned[all_scores_df_cleaned['generation'] == gen]['interface_score'], bins=bins)[0] for gen in unique_generations}

    bottom = np.zeros(len(bin_centers))
    for gen in unique_generations:
        ax.bar(bin_centers, counts[gen], bottom=bottom, width=np.diff(bins), label=f'Generation {gen}', color=color_map[gen], align='center')
        bottom += counts[gen]

    legend_elements = [Line2D([0], [0], marker='o', color='w', label=f'Generation {gen}',
                              markerfacecolor=color_map[gen], markersize=10) for gen in unique_generations]
    #ax.legend(handles=legend_elements, title="Generation")

    ax.set_title('Stacked Histogram of Interface Scores by Generation')
    ax.set_xlabel('Interface Score')
    ax.set_ylabel('Count')

    ax.set_xlim(-32.5, -13.5)

def plot_delta_scores():
    all_scores_df = pd.read_csv(ALL_SCORES_CSV)
    all_scores_df = all_scores_df.dropna(subset=['total_score'])
    
    # Calculate combined scores using normalized scores
    _, _, _, _, combined_scores = normalize_scores(all_scores_df, print_norm=True, norm_all=True)
    
    # Add combined scores to the DataFrame
    all_scores_df['combined_score'] = combined_scores

    # Calculate delta scores
    all_scores_df['delta_combined'] = all_scores_df.apply(lambda row: row['combined_score'] - all_scores_df.loc[all_scores_df['index'] == int(float(row['parent_index'])), 'combined_score'].values[0] if row['parent_index'] != "Parent" else 0, axis=1)
    all_scores_df['delta_total'] = all_scores_df.apply(lambda row: row['total_score'] - all_scores_df.loc[all_scores_df['index'] == int(float(row['parent_index'])), 'total_score'].values[0] if row['parent_index'] != "Parent" else 0, axis=1)
    all_scores_df['delta_interface'] = all_scores_df.apply(lambda row: row['interface_score'] - all_scores_df.loc[all_scores_df['index'] == int(float(row['parent_index'])), 'interface_score'].values[0] if row['parent_index'] != "Parent" else 0, axis=1)
    all_scores_df['delta_efield'] = all_scores_df.apply(lambda row: row['efield_score'] - all_scores_df.loc[all_scores_df['index'] == int(float(row['parent_index'])), 'efield_score'].values[0] if row['parent_index'] != "Parent" else 0, axis=1)

    fig, axs = plt.subplots(2, 2, figsize=(15, 10))

    def plot_violin(ax, delta_scores, title, all_scores_df):
        all_scores_df['tmp'] = delta_scores
        all_scores_df = all_scores_df.dropna(subset=['tmp'])

        max_gen = int(all_scores_df['generation'].max())
        generations = np.arange(0, max_gen + 1)
        violin_data = [all_scores_df[all_scores_df['generation'] == gen]['tmp'] for gen in generations]

        # Create violin plots
        parts = ax.violinplot(violin_data, positions=generations, showmeans=False, showmedians=True)

        # Customizing the color of violin plots
        for pc in parts['bodies']:
            pc.set_facecolor('green')
            pc.set_edgecolor('black')
            pc.set_alpha(0.7)

        # Customizing the color of the median lines
        for partname in ('cbars', 'cmins', 'cmaxes'):
            vp = parts.get(partname)
            if vp:
                vp.set_edgecolor('tomato')
                vp.set_linewidth(0.5)

        vp = parts.get('cmedians')
        if vp:
            vp.set_edgecolor('tomato')
            vp.set_linewidth(2.0)

        ax.set_title(title)
        ax.set_xlabel('Generation')
        ax.set_ylabel('Delta Score')
        ax.yaxis.grid(True, linestyle='--', which='major', color='grey', alpha=0.7)

    plot_violin(axs[0, 0], all_scores_df['delta_combined'], 'Delta Combined Score vs Generations', all_scores_df)
    plot_violin(axs[0, 1], all_scores_df['delta_total'], 'Delta Total Score vs Generations', all_scores_df)
    plot_violin(axs[1, 0], all_scores_df['delta_interface'], 'Delta Interface Score vs Generations', all_scores_df)
    plot_violin(axs[1, 1], all_scores_df['delta_efield'], 'Delta Efield Score vs Generations', all_scores_df)

    plt.tight_layout()
    plt.show()


def plot_tree_lin(leaf_nodes=None):
    all_scores_df = pd.read_csv(ALL_SCORES_CSV)
    _, _, _, _, combined_potentials = normalize_scores(all_scores_df, print_norm=False, norm_all=False, extension="potential")

    max_gen = int(all_scores_df['generation'].max())

    G = nx.DiGraph()

    for idx, row in all_scores_df.iterrows():
        G.add_node(idx, sequence=row['sequence'], interface_potential=row['interface_potential'], gen=int(row['generation']))
        if row['parent_index'] != "Parent":
            parent_idx = int(float(row['parent_index']))
            parent_sequence = all_scores_df.loc[all_scores_df.index == parent_idx, 'sequence'].values[0]
            current_sequence = row['sequence']
            # Calculate Hamming distance
            distance = hamming_distance(parent_sequence, current_sequence)
            # Add edge with Hamming distance as an attribute
            G.add_edge(parent_idx, idx, hamming_distance=distance)

    if leaf_nodes is not None:
        subgraph_nodes = set()
        for leaf in leaf_nodes:
            subgraph_nodes.update(nx.ancestors(G, leaf))
            subgraph_nodes.add(leaf)
        G = G.subgraph(subgraph_nodes)

    G_undirected = G.to_undirected()

    # Find connected components
    connected_components = list(nx.connected_components(G_undirected))

    largest_component = max(connected_components, key=len)
    # Create a subgraph of G using only the nodes in the largest component
    G_largest = G.subgraph(largest_component)

    def set_node_positions(G, node, pos, x, y, counts):
        pos[node] = (x, y)
        neighbors = list(G.successors(node))
        next_y = y - counts[node] / 2
        for neighbor in neighbors:
            set_node_positions(G, neighbor, pos, x + 1, next_y + counts[neighbor] / 2, counts)
            next_y += counts[neighbor]

    def count_descendants(G, node, counts):
        neighbors = list(G.successors(node))
        count = 1
        for neighbor in neighbors:
            count += count_descendants(G, neighbor, counts)
        counts[node] = count
        return count

    counts = {}
    root_node = list(largest_component)[0]
    count_descendants(G_largest, root_node, counts)

    pos = {}
    set_node_positions(G_largest, root_node, pos, 0, 0, counts)
    y_values = [y for x, y in pos.values()]
    y_span = max(y_values) - min(y_values)
    print(y_span)

    colors = combined_potentials
    colors[0] = np.nan
    normed_colors = [(x - np.nanmin(colors[1:])) / (np.nanmax(colors[1:]) - np.nanmin(colors[1:])) for x in colors]
    normed_colors = np.nan_to_num(normed_colors, nan=0)
    normed_colors = normed_colors**2

    # Convert positions to polar coordinates
    polar_pos = {node: ((x / (max(pos.values(), key=lambda p: p[0])[0] - min(pos.values(), key=lambda p: p[0])[0])) * 2 * np.pi, y) for node, (x, y) in pos.items()}

    # Convert polar coordinates to Cartesian coordinates for plotting
    cartesian_pos = {node: (radius * np.cos(angle), radius * np.sin(angle)) for node, (radius, angle) in polar_pos.items()}

    fig, ax = plt.subplots(figsize=(10, 10), dpi=300)

    # Draw the graph with the positions set
    for start, end in G_largest.edges():
        color = plt.cm.coolwarm_r(normed_colors[end])
        if float(normed_colors[end]) == 0.0:
            color = [0., 0., 0., 1.]
        linewidth = 0.1 + 2 * normed_colors[end] * 0.01

        x0, y0 = cartesian_pos[start]
        x1, y1 = cartesian_pos[end]
        ax.plot([x0, x1], [y0, y1], color=color, linewidth=linewidth)

    # Adjust axis labels and ticks for the swapped axes
    ax.axis('on')
    ax.set_title("Colored by Potential")
    ax.set_xlabel("X Coordinate")
    ax.set_ylabel("Y Coordinate")
    ax.set_yticks([])
    ax.set_xticks([])
    ax.axis('equal')
    ax.grid(False)
    plt.show()

def calculate_rank_order(matrix):
    # Calculate the occurrence frequency of each amino acid in each column
    unique, counts = np.unique(matrix, return_counts=True)
    frequencies = dict(zip(unique, counts))
    
    # Sort amino acids in each column by their frequency, then alphabetically
    sorted_amino_acids = sorted(frequencies.items(), key=lambda x: (x[1], -ord(x[0])), reverse=True)
    
    # Assign rank order based on sorted position
    rank_order = {amino_acid: rank for rank, (amino_acid, _) in enumerate(sorted_amino_acids, start=1)}
    
    # Replace amino acids with their rank order
    rank_matrix = np.vectorize(rank_order.get)(matrix)
    
    return rank_matrix

def seq_to_rank_order_matrix(sequences):
    # Convert sequences to a 2D numpy array (matrix) of characters
    matrix = np.array([list(seq) for seq in sequences])
    
    # Initialize an empty matrix to store the rank order numbers
    rank_order_matrix = np.zeros(matrix.shape, dtype=int)
    
    # Calculate rank order for each column
    for i in range(matrix.shape[1]):  # Iterate over columns
        column = matrix[:, i]
        rank_order_matrix[:, i] = calculate_rank_order(column)
    
    return rank_order_matrix

def seq_to_numeric(seq):
    # Define a mapping for all 20 standard amino acids plus 'X' for unknown
    mapping = {
        'A': 1,  'C': 2,  'D': 3,  'E': 4,
        'F': 5,  'G': 6,  'H': 7,  'I': 8,
        'K': 9,  'L': 10, 'M': 11, 'N': 12,
        'P': 13, 'Q': 14, 'R': 15, 'S': 16,
        'T': 17, 'V': 18, 'W': 19, 'Y': 20,
        'X': 0   # 'X' for any unknown or non-standard amino acid
    }
    numeric_seq = [mapping[char] for char in seq]
    return numeric_seq

def plot_pca_umap():
    
    all_scores_df = pd.read_csv(ALL_SCORES_CSV)

    all_scores_df = all_scores_df.dropna(subset=['total_score'])
    all_scores_df = all_scores_df.dropna(subset=['catalytic_score'])
    all_scores_df = all_scores_df.dropna(subset=['interface_score'])
    
    numeric_seqs = seq_to_rank_order_matrix(all_scores_df['sequence'].tolist())
    
    # Perform PCA
    pca = PCA(n_components=2)
    pca_result = pca.fit_transform(numeric_seqs)

    pca3 = PCA(n_components=3)
    pca_result3 = pca3.fit_transform(numeric_seqs)

    # Analyze PCA loadings for PC1
    # loadings = pca.components_.T[:, 0]  # Loadings for PC1
    # plt.figure(figsize=(10, 4))
    # plt.bar(range(len(loadings)), loadings)
    # plt.title('PCA Loadings for PC1')
    # plt.xlabel('Sequence Position')
    # plt.ylabel('Loading Value')
    # plt.show()

    # Perform UMAP
    reducer = umap.UMAP()
    umap_result = reducer.fit_transform(numeric_seqs)

    # Create a figure and a 2x2 grid of subplots
    fig, axs = plt.subplots(3, 1, figsize=(12, 24))  # Adjust the figure size as needed

    # Define a base font size
    base_font_size = 10  # Adjust here

    # Plot UMAP Interface score
    axs[0].scatter(umap_result[:, 0], umap_result[:, 1], c=all_scores_df['interface_score'], cmap='viridis', alpha=0.6, s=1)
    cbar = fig.colorbar(axs[0].collections[0], ax=axs[0], label='Interface Score')
    axs[0].set_title('UMAP of Sequences - Interface score', fontsize=base_font_size * 2)
    axs[0].set_xlabel('UMAP1', fontsize=base_font_size * 2)
    axs[0].set_ylabel('UMAP2', fontsize=base_font_size * 2)
    cbar.set_label('Interface Score', size=base_font_size * 2)

    # Filter the DataFrame to include only rows where 'total_score' is <= -340
    filtered_df = all_scores_df[all_scores_df['total_score'] <= -340]
    filtered_umap_result = umap_result[all_scores_df['total_score'] <= -340]

    # Now plot using the filtered data
    axs[1].scatter(filtered_umap_result[:, 0], filtered_umap_result[:, 1], c=filtered_df['total_score'], cmap='viridis', alpha=0.6, s=1)
    cbar = fig.colorbar(axs[1].collections[0], ax=axs[1], label='Total Score')
    axs[1].set_title('UMAP of Sequences - Total score', fontsize=base_font_size * 2)
    axs[1].set_xlabel('UMAP1', fontsize=base_font_size * 2)
    axs[1].set_ylabel('UMAP2', fontsize=base_font_size * 2)
    cbar.set_label('Total Score', size=base_font_size * 2)

    # Plot UMAP with 'index' as the color
    axs[2].scatter(umap_result[:, 0], umap_result[:, 1], c=all_scores_df['index'], cmap='viridis', alpha=0.6, s=1)
    cbar = fig.colorbar(axs[2].collections[0], ax=axs[2], label='Generation')
    axs[2].set_title('UMAP of Sequences - Generation', fontsize=base_font_size * 2)
    axs[2].set_xlabel('UMAP1', fontsize=base_font_size * 2)
    axs[2].set_ylabel('UMAP2', fontsize=base_font_size * 2)
    cbar.set_label('Generation', size=base_font_size * 2)

    plt.tight_layout()
    plt.show()



def plot_esm_umap():

    #ESM embeddings and UMAP
    def prepare_data(sequences):
        """ Convert a list of protein sequences to the model's input format. """
        batch_tokens = []
        for seq in sequences:
            tokens = torch.tensor([alphabet.encode(seq)], dtype=torch.long)
            batch_tokens.append(tokens)
        return torch.cat(batch_tokens)

    # 1. Load ESM model
    model, alphabet = esm.pretrained.esm2_t33_650M_UR50D()
    model.eval()

    # Load and preprocess data
    all_scores_df = pd.read_csv(ALL_SCORES_CSV)
    all_scores_df.dropna(subset=['total_score', 'catalytic_score', 'interface_score', 'sequence'], inplace=True)

    # Extract sequences
    sequences = all_scores_df['sequence'].tolist()

    with torch.no_grad():
        tokens = prepare_data(sequences)
        results = model(tokens, repr_layers=[33])  # Specify the layer you want
        token_embeddings = results["representations"][33]

        # Mean pooling over positions
        sequence_embeddings = token_embeddings.mean(dim=1)
        
    embeddings_array = sequence_embeddings.cpu().numpy()

    # Perform UMAP
    reducer = umap.UMAP()
    umap_result = reducer.fit_transform(embeddings_array)

    # Create a figure and a 2x2 grid of subplots
    fig, axs = plt.subplots(3, 1, figsize=(12, 24))  # Adjust the figure size as needed

    # Define a base font size
    base_font_size = 10  # Adjust here

    # Plot UMAP Interface score
    axs[0].scatter(umap_result[:, 0], umap_result[:, 1], c=all_scores_df['interface_score'], cmap='viridis', alpha=0.6, s=1)
    cbar = fig.colorbar(axs[0].collections[0], ax=axs[0], label='Interface Score')
    axs[0].set_title('UMAP of Sequences - Interface score', fontsize=base_font_size * 2)
    axs[0].set_xlabel('UMAP1', fontsize=base_font_size * 2)
    axs[0].set_ylabel('UMAP2', fontsize=base_font_size * 2)
    cbar.set_label('Interface Score', size=base_font_size * 2)

    # Filter the DataFrame to include only rows where 'total_score' is <= -340
    filtered_df = all_scores_df[all_scores_df['total_score'] <= -340]
    filtered_umap_result = umap_result[all_scores_df['total_score'] <= -340]

    # Now plot using the filtered data
    axs[1].scatter(filtered_umap_result[:, 0], filtered_umap_result[:, 1], c=filtered_df['total_score'], cmap='viridis', alpha=0.6, s=1)
    cbar = fig.colorbar(axs[1].collections[0], ax=axs[1], label='Total Score')
    axs[1].set_title('UMAP of Sequences - Total score', fontsize=base_font_size * 2)
    axs[1].set_xlabel('UMAP1', fontsize=base_font_size * 2)
    axs[1].set_ylabel('UMAP2', fontsize=base_font_size * 2)
    cbar.set_label('Total Score', size=base_font_size * 2)

    # Plot UMAP with 'index' as the color
    axs[2].scatter(umap_result[:, 0], umap_result[:, 1], c=all_scores_df['index'], cmap='viridis', alpha=0.6, s=1)
    cbar = fig.colorbar(axs[2].collections[0], ax=axs[2], label='Generation')
    axs[2].set_title('UMAP of Sequences - Generation', fontsize=base_font_size * 2)
    axs[2].set_xlabel('UMAP1', fontsize=base_font_size * 2)
    axs[2].set_ylabel('UMAP2', fontsize=base_font_size * 2)
    cbar.set_label('Generation', size=base_font_size * 2)

    plt.tight_layout()
    plt.show()

def find_mutations(seq1, seq2):
    # Function to compare sequences and find mutation positions
    return [i for i, (a, b) in enumerate(zip(seq1, seq2)) if a != b]

def normalize_columnwise(matrix):
    min_vals = matrix.min(axis=0)
    max_vals = matrix.max(axis=0)
    # Avoid division by zero
    denom = np.where((max_vals - min_vals) == 0, 1, (max_vals - min_vals))
    normalized_matrix = (matrix - min_vals) / denom
    return normalized_matrix

def plot_mut_location():
    # Load the data
    all_scores_df = pd.read_csv(ALL_SCORES_CSV)

    all_scores_df = all_scores_df.dropna(subset=['sequence'])

    # Assuming the maximum length of sequences is 125
    max_length = 125
    max_generation = int(all_scores_df['generation'].max())

    # Initialize a matrix to hold mutation frequencies
    mutation_matrix = np.zeros((max_length, max_generation + 1))

    # Populate the mutation matrix
    for _, row in all_scores_df.iterrows():
        if pd.notnull(row['parent_index']) and row['parent_index'] != "Parent":  # Check if there's a valid parent
            parent_seq = all_scores_df.loc[all_scores_df['index'] == float(row['parent_index']), 'sequence'].values[0]
            mutations = find_mutations(row['sequence'], parent_seq)
            for pos in mutations:
                mutation_matrix[pos, int(row['generation'])] += 1

    # Normalize the mutation_matrix column-wise (i.e., each generation separately)
    normalized_mutation_matrix = normalize_columnwise(mutation_matrix)

   # Plotting
    fig, ax = plt.subplots(figsize=(10, 6))
    c = ax.imshow(normalized_mutation_matrix, aspect='auto', origin='lower', cmap='viridis', extent=[0, max_generation, 0, max_length])
    ax.set_xlabel('Generation')
    ax.set_ylabel('Position along AA chain')
    ax.set_title('Frequency of Mutation Over Generations')
    fig.colorbar(c, ax=ax, label='Normalized Frequency of Mutation')
    plt.show()



"""
Calculates Redox potential using the BioDC software.

Functions:
    - prepare_BioDC: Prepares commands for BioDC job submission.
"""

import os
import shutil

def prepare_BioDC(self, 
                       index, 
                       cmd):
    """
    Redox potential calculation for {index} with BioDC
    
    Parameters:
    index (str): The index of the protein variant to be analyzed.
    cmd (str): Growing list of commands to be exected by run_design using submit_job.

    Returns:
    cmd (str): Command to be exected by run_design using submit_job.
    """

    
    filename = f'{self.FOLDER_DESIGN}/{index}'
    filename_in = self.all_scores_df.at[index, "final_variant"]
    variant_name = os.path.basename(filename_in)
    # Make directories
    os.makedirs(f"{filename}/scripts", exist_ok=True)
    os.makedirs(f"{filename}/BioDC", exist_ok=True)
    os.makedirs(f"{filename}/BioDC/{variant_name}", exist_ok=True)

    with open(f"{filename_in}.pdb", "r") as f:
        lines = f.readlines()
    
    with open(f"{filename}/BioDC/{variant_name}/{variant_name}.pdb", "w") as f:
        for line in lines:
            if not (line.startswith("ATOM") or line.startswith("HETATM") or line.startswith("TER")): continue
            if "OXT" in line: continue
            if line.split()[-1] == "H": continue
            if "HEM X   1" in line:
                line = line.replace("HEM X   1", f"HEM B {int(self.HEME_RESI[2])}")
                line = line.replace("HETATM", "ATOM  ")
            f.write(line)
            
    input_file = f"""DivSel = 1                     
OriginalPDB = {variant_name}
SelDisulfides = no
ChooseMut = no
SelCpH = yes
SelectAllTitratable = yes
CreateResIndexingMethod = auto
ProcessPDBChoice = yes
RedoxState_1 = R
OutPrefix = BioDC
SolvEnv = imp
StructRelaxCompChoice = S
NProc = 1
continue_to_div_2 = yes
use_existing_min = yes
confirm_single_heme = yes
calculator_selection = 1
protein_dielectric = 5.5 
solvent_dielectric = 78.2
dg_method = 1
ref_state = red
use_eps_113 = yes
pbsa_type = 1
use_membrane = no
run_parallel = no
istrng_113 = 150
continue_calculations = no
continue_to_div_3 = no
"""
    with open(f"{filename}/BioDC/{variant_name}/input.txt", "w") as f:
        f.write(input_file)

    cmd += f"""### BioDC ####

cd BioDC/{variant_name}

module load vmd/1.9.3 amber/24 gcc/12 impi/2021.9

echo "Starting Xvfb..."
Xvfb :99 -screen 0 1024x768x16 &
export DISPLAY=:99
export VMDFORCECPU=1 
sleep 2
biodc-cli
kill %1

cd ../..
"""
    
    if self.REMOVE_TMP_FILES:
        cmd += f"""
# Removing temporary directory
rm -r BioDC
"""
        
    return cmd

def get_redox_score(self, index, score_type):
  
    with open(f"{self.FOLDER_DESIGN}/{int(index)}/BioDC/{self.WT}_{score_type}_{index}/EE/DG.txt", "r") as f:
        redoxpotential = f.read()
    redoxpotential = float(redoxpotential.split()[-2])
    redox_score = (self.TARGET_REDOX-redoxpotential)**2
    return redox_score, redoxpotential

def BioDC_check(self):
    """
    Checks if BioDC is installed
    """
    if not os.path.exists(self.FOLDER_BioDC):
        logging.error(f"BioDC not installed in {self.FOLDER_BioDC}.")
        logging.error(f"Install using: 'cd {os.path.dirname(self.FOLDER_BioDC)} && \
        git clone -b dev_v3.2 https://github.com/Mag14011/BioDC.git && \
        cd BioDC && \
        pip uninstall biodc  && \
        pip install . ")
        sys.exit()